#!/usr/bin/env python

"""Discrete State Markov Model Simulator.

This module implements a discrete trait simulator based on a Markov
model. The root state is either randomly sampled, or assigned, and
transitions among states occur probabilistiically along the edges of
a phylogeny. Transition rates can be specified for models with
unequal transition rates.   

References
----------
- Yang...
"""

from typing import Optional, List, Any, Union
from dataclasses import dataclass, field
from enum import Enum
import numpy as np
import pandas as pd
import scipy.linalg
from toytree import ToyTree
from toytree.core.apis import add_subpackage_method, PhyloCompAPI

__all__ = [
    "get_markov_model",
    "simulate_discrete_data",
]


class ModelType(Enum):
    """Supported named Markov model types to be fit or simulated.

    This will raise an exception is user tries to enter a value not
    supported in this class.
    """
    ER = "ER"
    SYM = "SYM"
    ARD = "ARD"


@dataclass
class MarkovModel:
    """Generate a Q-matrix under ER, SYM, or ARD models.

    This is intended primarily for internal use by toytree.pcm.

    Examples
    --------
    >>> model = MarkovModel(3, "ARD")
    >>> print(model.qmatrix)
    """
    nstates: int
    """: Number of possible character states."""
    mtype: ModelType
    """: Model type ("ER", "SYM", "ARD")."""
    relative_rates: Optional[np.ndarray] = None
    """: Relative transition rates. If not entered then values are
    sampled within model constraints given the random seed."""
    state_frequencies: Optional[np.ndarray] = None
    """: Equilibrium frequencies of the nstates."""
    rate_scalar: float = 1.0
    """: Rate scalar to multiple relative rates by."""
    seed: Optional[int] = None
    """: Random seed used if relative_rates is None."""

    # attributes filled after init.
    rng: np.random.Generator = field(init=False, repr=False)
    """: Random number generator init with seed."""
    transition_matrix: np.ndarray = field(init=False)
    """: Transition probability matrix (P)."""
    qmatrix: np.ndarray = field(init=False)
    """: Instantaneous rate matrix (Q)."""

    def __post_init__(self):
        self.rng = np.random.default_rng(self.seed)
        self.mtype = ModelType(self.mtype)
        self._check_rates()
        self._check_freqs()
        self._set_transition_matrix()
        self._set_qmatrix()

    def _check_rates(self):
        """Checks the relative rates matrix given mtype and nstates.

        If a user entered the matrix it is checked to be appropriate
        given the model type. If no matrix is entered then a random
        matrix is generated that is appropriate for the model type.

        Examples
        --------
        >>> MarkovModel(2, "ER").relative_rates
        [[0, 1],[1, 0]]
        """
        # user entered rate matrix
        rates = self.relative_rates
        
        # if no user-entered rates then sample random rates constrained
        # by the model type.
        if rates is None:
            if self.mtype.name == "SYM":
                rates = np.random.uniform(0.5, 2, (self.nstates, self.nstates))
                rates[0, 1] = 1
                lower = np.tril_indices_from(rates)
                upper = np.tril_indices_from(rates)[::-1]
                rates[lower] = rates[upper]
            elif self.mtype.name == "ARD":
                rates = np.random.uniform(0.5, 2, (self.nstates, self.nstates))
                rates[0, 1] = 1
            else:
                rates = np.ones((self.nstates, self.nstates))
            np.fill_diagonal(rates, 0)
        
        # if user entered rates then check that they are valid.
        else:
            rates = np.array(rates)
            # check if singular for ER model
            if self.mtype.name == "ER":
                if rates.size == 1:
                    rates = (
                        np.repeat(rates, self.nstates * self.nstates)
                        .reshape((self.nstates, self.nstates))
                    )
                assert len(set(rates[rates != 0])) == 1, (
                    "all rates should be equal in ER model. See SYM model.")
            # check if symmetric for SYM models
            elif self.mtype.name == "SYM":
                assert np.allclose(rates, rates.T, rtol=1e-5, atol=1e-8), (
                    "rates should be symmetric in SYM model. See ARD model.")
            # check shape
            assert rates.shape == (self.nstates, self.nstates), (
                f"given nstates={self.nstates} the rates matrix should "
                f"be shape ({self.nstates}, {self.nstates}).")
            np.fill_diagonal(rates, 0)
        self.relative_rates = rates

    def _check_freqs(self):
        """Checks stationary frequencies array given model and nstates.

        Entered values are checked for correct size and that they
        sum to one. If no values are entered then a uniform frequency
        of the correct size is sampled for ER, or a random sample that
        sums to one is sampled for SYM and ARD.

        Examples
        --------
        >>> MarkovModel(3, "ER").state_frequencies
        [1/3, 1/3, 1/3]
        """
        freqs = self.state_frequencies
        if freqs is None:
            if self.mtype.name in ("SYM", "ARD"):
                freqs = np.random.uniform(0.5, 2, self.nstates)
                freqs /= freqs.sum()
            else:
                freqs = np.repeat(1.0 / self.nstates, self.nstates)
        else:
            freqs = np.array(freqs)
            if self.mtype.name in ("SYM", "ARD"):
                assert freqs.size == self.nstates, (
                    f"states_frequencies should be len={self.nstates}.")
            else:
                fixed = np.repeat(1.0 / self.nstates, self.nstates)
                assert np.allclose(freqs, fixed), (
                    f"ER model with nstates={self.nstates} has fixed state "
                    f"frequences: {fixed}. See SYM or ARD models.")
                freqs = fixed
        assert np.allclose(sum(freqs), 1.0), (
            f"state_frequencies must sum to 1. You entered: {freqs}")
        self.state_frequencies = freqs

    def _set_transition_matrix(self):
        """Sets the transition probabilities (freqs.T x rates).

        Examples
        --------
        >>> MarkovModel(3, "ER").transition_matrix
        [[0, 1/3, 1/3, 1/3]
         [1/3, 0, 1/3, 1/3]
         [1/3, 1/3, 0, 1/3]
         [1/3, 1/3, 1/3, 0]]
        """
        # matrix multiply and set diagonal back to zero.
        sfreq_mat = np.eye(self.nstates) * self.state_frequencies
        trans_mat = np.matmul(sfreq_mat, self.relative_rates)
        np.fill_diagonal(trans_mat, 0)

        # divide by scaling factor (largest rowsum excluding diagonal)
        scaling = np.sum(trans_mat, axis=1).max()
        trans_mat /= scaling

        # fill diagonals to sum to 1
        for idx in range(trans_mat.shape[0]):
            trans_mat[idx, idx] = abs(1 - trans_mat[idx].sum())
        self.transition_matrix = self.rate_scalar * trans_mat

    def _set_qmatrix(self):
        """Set a instantaneous transition probability matrix (Q).

        This matrix is the probability of transitioning from one
        state to another in a single unit of time. The rates are
        normalized by a matrix scaling factor, calculated as the max
        rowsum of the state_frequencies matrix x rates_matrix.
        >>> T = (rates_matrix * state_frequencies) * rate / scaling
        >>> Q = (set T diagonal so rows sum to 0.)

        Parameters
        ----------
        rate: float
            A scalar by which the Q-matrix will be multipled to act
            as a unit scaler. Example, rate=1e-6 would mean that a 1
            in the relatives rates matrix represents 1 change per
            million edge length units on a tree.
        rates_matrix: numpy.ndarray
            The relative transition rates between states as an array
            of size (nstates x nstates). Values on the diagonal are
            ignored. Only relative differences matter. See rate.
        state_frequencies: numpy.ndarray
            Equilibrium frequencies of states 0-n in order (must sum
            to one.

        Examples
        --------
        >>> MarkovModel(nstates=3, model="ER").qmatrix
        [[-1, 1/3, 1/3, 1/3]
         [1/3, -1, 1/3, 1/3]
         [1/3, 1/3, -1, 1/3]
         [1/3, 1/3, 1/3, -1]]
        """
        self.qmatrix = self.transition_matrix - (self.rate_scalar * np.eye(self.nstates))

    def get_transition_probability_matrix(self, time: float) -> np.ndarray:
        """Return a transition probability matrix for a length of time.

        This represents the probability that over the length of time
        a character starting in one state will transition to another:
        >>> T = (rates_matrix * state_frequencies) / scaling_factor
        >>> Q = (set diagonal of T to sum rows to 0)
        >>> P(t) = expm(Q * rate * time)

        Parameters
        ----------
        time: float
            Length of time over which state transitions can occur.
            Note that the rates in the rate_matrix should correspond
            to the same units as time (i.e. transitions / unit time).

        Examples
        --------
        >>> mod = MarkovModel(nstates=3, model="ER")
        >>> mod.get_transition_probability_matrix(time=1.)
        [[0.44769785 0.18410072 0.18410072 0.18410072]
         [0.18410072 0.44769785 0.18410072 0.18410072]
         [0.18410072 0.18410072 0.44769785 0.18410072]
         [0.18410072 0.18410072 0.18410072 0.44769785]]

        Over zero amount of time no change is expected:
        >>> mod.get_transition_probability_matrix(time=0)
        [[1. 0. 0. 0.]
         [0. 1. 0. 0.]
         [0. 0. 1. 0.]
         [0. 0. 0. 1.]]

        Over very long time scales transition probabilities will match
        the state_frequencies (1 / nstates for ER model):
        >>> mod.get_transition_probability_matrix(time=1000)
        [[0.25 0.25 0.25 0.25]
         [0.25 0.25 0.25 0.25]
         [0.25 0.25 0.25 0.25]
         [0.25 0.25 0.25 0.25]]
        """
        return scipy.linalg.expm(self.qmatrix * time)

    # def _repr_html_(self):
    #     """Return a html representation of the Markov model.
    #     TODO: return multiple tables?
    #     """

    def __repr__(self):
        """Return a str representation of the Markov model."""
        return f"MarkovModel(nstates={self.nstates}, model={self.mtype.name})"
        # ,\nT matrix\n{self.transition_matrix}\nQ matrix\n{self.qmatrix}\nTransition Probabilities (t=1)\n{self.get_transition_probability_matrix(time=1)}"



@dataclass
class DiscreteMarkovSimulator:
    """Simulate a discrete trait on a tree given a Q-matrix.

    This is intended primarily for internal use by toytree.pcm,
    with the user-facing functions available from factory functions
    like :meth:`simulate_discrete_data`. This takes as input a
    ToyTree and MarkModel instances
    """
    tree: ToyTree
    """: ToyTree with edge lengths in units of ..."""
    model: MarkovModel
    """: MarkovModel object with parameterized Q matrix."""
    root_state: Optional[int] = None
    """: Integer character state at the root."""
    seed: Optional[int] = None
    """: ..."""
    rng: np.random.Generator = field(init=False)
    """: ..."""

    def __post_init__(self):
        self.rng = np.random.default_rng(self.seed)

    def _edge_sim(self, state: int, time: float) -> int:
        """Return the state at end of this time given starting state."""
        prob = scipy.linalg.expm(self.model.qmatrix * time)
        return self.rng.multinomial(1, prob[state]).argmax()

    def _traversal_sim(self) -> np.ndarray:
        """Traverse tree from root to tips simulating trait."""
        arr = np.zeros(self.tree.nnodes, dtype=np.int64)
        # sample a random root state
        if self.root_state is None:
            arr[-1] = self.rng.multinomial(1, self.model.state_frequencies).argmax()
        else:
            arr[-1] = self.root_state

        # ...
        for node in self.tree[::-1][1:]:
            parent_state = arr[node.up._idx]
            state = self._edge_sim(parent_state, node._dist)
            arr[node._idx] = state
        return arr

    def run(self, nreplicates: int) -> pd.DataFrame:
        """Return a DataFrame with simulated discrete (integer) data.

        Node idx labels as index, and columns labeled t0-t{replicates}.
        """
        arr = np.zeros((self.tree.nnodes, nreplicates), dtype=np.int64)
        for ridx in range(nreplicates):
            arr[:, ridx] = self._traversal_sim()
        traits = [f"t{i}" for i in range(nreplicates)]
        return pd.DataFrame(data=arr, index=range(self.tree.nnodes), columns=traits)


####################################################################
# API Exposed functions
####################################################################
def get_markov_model(
    nstates: int,
    model: str = "ER",
    rate_scalar: float = 1.0,
    relative_rates: Optional[np.ndarray] = None,
    state_frequencies: Optional[np.ndarray] = None,
    seed: Optional[int] = None,
) -> MarkovModel:
    """Return a parameterized MarkovModel instance.

    The MarkovModel class is used to get an instantaneous transition
    rate matrix (Q) for calculating the probabilities of transitions
    between discrete character states in a Markov model. This
    is used primarily for didactic purposes, and is also used
    internally in functions such as :meth:`~toytree.pcm.simulate_discrete_data`.

    It checks that the user input for rates and state_frequencies
    if valid given the model type and number of states, and can
    return random valid paramterizations for each model type.

    Parameters
    ----------
    nstates: int
        The number of states to simulate. States will be names as
        integers 0-nstates, unless you use `state_names` to rename.
    model: str
        The Markov model name ("ER", "SYM", or "ARD"). This is used
        to either sample random valid parameters for a model of the
        specified type, or to check that user-entered value are valid.
    rate_scalar: float
        A scalar by which the Q-matrix will be multipled to act
        as a unit scaler. Example, rate=1e-6 would mean that a 1
        in the relatives rates matrix represents 1 change per
        million edge length units on a tree.
    relative_rates: Optional[numpy.ndarray]
        The relative transition rates between states as an array
        of size (nstates x nstates). Values on the diagonal are
        ignored. Only relative differences matter. See rate.
    state_frequencies: Optional[numpy.ndarray]
        Equilibrium frequencies of states 0-n in order (must sum
        to one.
    seed: Optional[int]
        Integer seed for numpy random number generator.

    Returns
    -------
    MarkovModel
        A parameterized MarkovModel class instance. The parameters
        of Markov model can be accessed from this instance from its
        attributes `.qmatrix`, `.state_frequencies`, etc.

    Examples
    --------
    >>> print(toytree.pcm.get_markov_model(nstates=3, model="ER"))
    >>> print(toytree.pcm.get_markov_model(nstates=3, model="SYM"))
    >>> print(toytree.pcm.get_markov_model(nstates=3, model="ARD")
    """
    model = MarkovModel(
        mtype=str(model).upper(),
        nstates=nstates,
        rate_scalar=rate_scalar,
        relative_rates=relative_rates,
        state_frequencies=state_frequencies,
        seed=seed,
    )
    return model


@add_subpackage_method(PhyloCompAPI)
def simulate_discrete_data(
    tree: ToyTree,
    nstates: int,
    model: str = "ER",
    relative_rates: Optional[np.ndarray] = None,
    state_frequencies: Optional[np.ndarray] = None,
    root_state: Optional[int] = None,
    rate_scalar: Optional[float] = 1.0,
    nreplicates: int = 1,
    tips_only: bool = False,
    state_names: Optional[List[Any]] = None,
    seed: Optional[int] = None,
    inplace: bool = False,
) -> Union[pd.DataFrame, None]:
    """Return trait values simulated under a discrete Markov model.

    The number of states and model type can be entered without any
    parameters to the Markov model (e.g., relative rates and/or
    state_frequencies) to generate a random set of parameters that
    are valid under the specified model. Or, if parameters are entered
    then they are checked for validity with the specified model type,
    and then used to parameterize the Markov model simulation.

    Parameters
    ----------
    tree: toytree.ToyTree
        The tree on which to simulate the trait, usually ultrametric.
    nstates: int
        The number of states to simulate. States will be names as
        integers 0-nstates, unless you use `state_names` to rename.
    model: str
        The Markov model name ("ER", "SYM", or "ARD"). This is used
        to either sample random valid parameters for a model of the
        specified type, or to check that user-entered value are valid.
    relative_rates: Optional[numpy.ndarray]
        The relative transition rates between states as an array
        of size (nstates x nstates). Values on the diagonal are
        ignored. Only relative differences matter. See rate_scalar.
    state_frequencies: Optional[numpy.ndarray]
        Equilibrium frequencies of states 0-n in order (must sum
        to one.
    root_state: Optional[int]
        The state at the root node (start of simulation). If None a
        state is randomly uniformly sampled from nstates.
    rate_scalar: float
        A scalar by which the Q-matrix will be multipled to act
        as a unit scaler. Example, rate=1e-6 would mean that a 1
        in the relatives rates matrix represents 1 change per
        million edge length units on a tree. Default=1 (no scaling)
    seed: Optional[int]
        Integer seed for numpy random number generator.
    tips_only: bool
        If True values are only returned for tip Nodes, else values are
        returned for all Nodes in the tree.
    state_names: Optional[List[Any]]
        State names will be substituted for simulated discrete integer
        states in the entered order.
    nreplicates: int
        If nreplicates is > 0 then the returned object is a DataFrame
        with replicate numbers as columns labels, otherwise a single
        Series is returned.
    inplace: bool
        If False a DataFrame is returned, if True a ToyTree is returned
        with data stored to Nodes.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10)
    >>>
    >>> # simulate a single trait
    >>> toytree.pcm.simulate_discrete_data(tree, 2, "ER")
    >>> # 0    1
    >>> # 1    0
    >>> # 2    1
    >>> # 3    0
    >>> # 4    0
    >>> # 5    0
    >>> # Name: trait, dtype: int64
    >>>
    >>> # simulate multiple replicate simulations
    >>> toytree.pcm.simulate_discrete_data(
    >>>     tree=tree, nstates=2, model="ARD",
    >>>     rate=0.1,
    >>>     relative_rates=[[0, 2], [1, 0]],
    >>>     state_frequencies=[0.4, 0.6],
    >>>     nreplicates=5,
    >>>     tips_only=True,
    >>>     state_names=["A", "B"],
    >>> )
    >>>
    >>> #     t0  t1  t2  t3  t4
    >>> # 0   A   B   A   A   B
    >>> # 1   B   B   A   B   B
    >>> # 2   A   B   B   A   B
    >>> # 3   A   B   B   B   B
    >>> # 4   B   B   A   B   B
    >>> # ...
    """
    model = MarkovModel(
        mtype=model,
        nstates=nstates,
        relative_rates=relative_rates,
        state_frequencies=state_frequencies,
        rate_scalar=rate_scalar,
    )
    simulator = DiscreteMarkovSimulator(
        tree=tree,
        model=model,
        root_state=root_state,
        seed=seed,
    )

    # get DataFrame or Series result
    traits = simulator.run(nreplicates)

    # subsample to return only tips
    if tips_only:
        traits = traits.iloc[:tree.ntips]

    # rename data from numeric to user defined
    if state_names is not None:
        for idx, value in enumerate(state_names):
            traits.replace(to_replace=idx, value=value, inplace=True)

    # store data to ToyTree or return as DataFrame
    if not inplace:
        return traits
    tree.set_node_data_from_dataframe(traits, inplace=True)
    return None


# def draw_markov_model():
#     """Returns a toyplot graph representation of a Markov model.
#     """
#     # c, a, m = toyplot.graph()
#     # return c, a, m


if __name__ == "__main__":

    import toytree

    # get a single tree with trait values
    tre = toytree.rtree.unittree(10, treeheight=10, seed=123)
    data = simulate_discrete_data(
        tree=tre,
        nstates=3,
        model="SYM",
        rate_scalar=1.,
        state_frequencies=[0.1, 0.3, 0.6],
        tips_only=True,
        root_state=0,
        nreplicates=10,
        state_names=['a', 'b', 'c'],
    )
    # print(data.T)
    # data = simulate_discrete_data(tre, nstates=2, inplace=True)
    print(data)

    # get many trees with traits simulated
    # ...

    # get a dataframe of tip or node traits
    # ...

    # get an array of many replicate sim data
    # ...

    model = get_markov_model(model="SYM", nstates=3, state_frequencies=[0.1, 0.2, 0.7], rate_scalar=0.1)
    print(model)
    print(model.get_transition_probability_matrix(time=10))



#!/usr/bin/env python

"""Parse a ToyTree object from an edge table.

Example
-------
>>> import toyplot
>>> import numpy as np
>>> import scipy.spatial
>>> import scipy.cluster

# generate 100 random data points for 12 things
>>> rng = np.random.default_rng(123)
>>> data = rng.random(size=(6, 100))

# calculate a measure of distance between samples
>>> dists = scipy.spatial.distance.pdist(data)

# infer a hierarchical tree by clustering samples by distances
>>> edges = scipy.cluster.hierarchy.average(dists)

# draw the tree
>>> canvas, axes, mark = tree.draw()
"""

import numpy as np
import toytree


def _build_tree_from_scipy_dist_table(table: np.ndarray):
    """Return ToyTree from a scipy distance table.
    
    A greedy algorithm to implement hierarchical clustering of samples
    based on distances in a distance table.
    """
    name_to_node = {}
    
    # iterate over rows of the table
    for ridx in range(table.shape[0]):
        
        # internal name is row number + ntips
        iname = int(ridx + table[-1, -1])
        
        # create new internal node
        internal = toytree.Node(name=str(iname))
        name_to_node[iname] = internal
        
        # iterate over child indices
        for cidx in table[ridx, [0, 1]]:
            cidx = int(cidx)
            
            # create child node if it doesn't exist
            if cidx not in name_to_node:
                child = toytree.Node(name=str(cidx))
                name_to_node[cidx] = child
            else:
                child = name_to_node[cidx]
                
            # set distance
            child.dist = table[ridx, 2]
            
            # connect to parent
            child.up = internal
            internal.children.append(child)
    return toytree.ToyTree(internal)


if __name__ == "__main__":

    import toyplot
    import numpy as np
    import scipy.spatial
    import scipy.cluster

    # generate 100 random data points for 12 things
    rng = np.random.default_rng(123)
    data = rng.random(size=(6, 100))

    # calculate a measure of distance between samples
    dists = scipy.spatial.distance.pdist(data)

    # infer a hierarchical tree by clustering samples by distances
    edges = scipy.cluster.hierarchy.average(dists)
    print(edges)
    # draw the tree
    # canvas, axes, mark = tree.draw()


#!/usr/bin/env python

"""Phylogenetic independent contrasts for continuous traits.

References
----------
- Felsenstein 1985
- Harmon textbook
- ...

"""

from typing import Tuple


def get_phylogenetic_independent_contrasts(tree, feature):
    """Return a dictionary of {idx: standardized-contrasts}.

    Independent contrasts are calculated for every internal node
    of a tree for a selected continuous feature (trait) 
    under a Brownian motion model of evolution.

    Modified from IVY interactive (https://github.com/rhr/ivy/)

    Parameters
    ----------
    feature: (str)
        The name of a feature of the tree that has been mapped to all 
        tip nodes of the tree. 

    Returns
    -------
    Dict[int, Tuple(float, float, float, float)]

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=10, treeheight=1)
    >>> tree.set_node_data("trait", dict(range(10), range(10)), inplace=True)
    >>> pics = toytree.pcm.get_phylogenetic_independent_contrasts(tree, "trait")
    >>> ...
    """
    # get current node features at the tips
    tips = [tree[i] for i in range(tree.ntips)]
    fdict = {i.name: getattr(i, feature) for i in tips}
    data = {i: fdict[i] for i in fdict if i in tree.get_tip_labels()}

    # apply dynamic function from ivy to return dict results
    results = _dynamic_pic(tree.treenode, data, results={})

    # return dictionary mapping nodes to (mean, var, contrast, cvar)
    return results


### NEEDS MORE...

# def get_continuous_ancestral_states(tre, feature):
#     """Infer ancestral states on ancestral nodes for continuous traits
#     under a brownian motion model of evolution.

#     Modified from IVY interactive (https://github.com/rhr/ivy/)   

#     Returns:
#     --------
#     toytree (toytree.ToyTree)
#         A modified copy of the input tree is returned with the mean 
#         ancestral value for the selected feature applied to all nodes 
#         of the tree. 
#     """
#     ntre = tre.copy()
#     resdict = get_phylogenetic_independent_contrasts(ntre, feature)
#     ntre = ntre.set_node_data(
#         feature, 
#         {i.idx: j[0] for (i, j) in resdict.items()}
#     )
#     return ntre


def _dynamic_pic(node, data, results):
    """Used internally by get_independent_contrasts.

    Phylogenetic independent contrasts. Recursively calculate 
    independent contrasts of a bifurcating node given a dictionary
    of trait values.

    Modified from IVY interactive (https://github.com/rhr/ivy/)

    Parameters
    ----------
    node: 
        A node object
    data: dict
        Mapping of leaf names to character values

    Returns
    -------
    dict
        Mapping of internal nodes to tuples containing ancestral
        state, its variance (error), the contrast, and the
        contrasts's variance.

    TODO: modify to accommodate polytomies.
    """
    # store in lists to support flexible number of children (e.g. polytomies)
    means = []
    variances = []

    # recursively does children until X and v are full
    for child in node.children:

        # child has children, do those first
        if child.children:

            # update results dict with children values
            _dynamic_pic(child, data, results)
            child_results = results[child]

            # store childrens values
            means.append(child_results[0])
            variances.append(child_results[1])

        # no child of child, so just do child
        else:
            means.append(data[child.name])
            variances.append(child.dist)

    # Xi - Xj is the contrast value
    means_i, means_j = means

    # vi + vj is the contrast variance
    vars_i, vars_j = variances

    # Xk is the reconstructed state at the node
    means_k = (
        ((1.0 / vars_i) * means_i + (1 / vars_j) * means_j) / 
        (1.0 / vars_i + 1.0 / vars_j)
    )

    # vk is the variance
    vars_k = node.dist + (vars_i * vars_j) / (vars_i + vars_j)

    # store in dictionary and return
    results[node] = (means_k, vars_k, means_i - means_j, vars_i + vars_j)
    return results


# def independent_contrasts(tre, feature):
#     """
#     Set independent contrasts as features on internal nodes labeled
#     as ...
#     """
#     ntre = tre.copy()
#     resdict = PIC(ntre, feature)
#     ntre = ntre.set_node_values(
#         feature="{}-contrast",
#         values={i.name: j[2] for (i, j) in resdict.items()}
#     )
#     ntre = ntre.set_node_values(
#         feature="{}-contrast-var",
#         values={i.name: j[3] for (i, j) in resdict.items()}
#     )        
#     return ntree

# single test
if __name__ == "__main__":

    import toyplot
    import toytree

    CMAP = toyplot.color.brewer.map("BlueRed", reverse=True)

    TREE = toytree.rtree.imbtree(5, 1e6)
    TREE = TREE.set_node_data(
        "g", 
        mapping={i: 5 for i in (2, 3, 4)},
        default=1,
    )

    TREE.draw(
        ts='p', 
        node_labels=TREE.get_node_data("g"),
        node_colors=[
            CMAP.colors(i, 0, 5) for i in TREE.get_node_data('g')]
        )

    # apply reconstruction
    ntree = get_phylogenetic_independent_contrasts(TREE, "g")

    # # new values are stored as -mean, -var, -contrasts, ...
    # evals = ntree.get_edge_values("g-mean")

    # # draw new tree
    # ntree.draw(
    #     ts='p', 
    #     node_labels=ntree.get_node_values("g-mean", 1, 1),
    #     node_colors=[
    #         colormap.colors(i, 0, 5) for i in 
    #         ntree.get_node_values('g-mean', 1, 1)]
    # )



#!/usr/bin/env python

"""Parse user args to `draw` and return drawing.

The `draw_toytree` func does the following:
    - calls `get_tree_style_updated_by_draw_args` to get apply user
      args to `draw` to a starting Treestyle. This includes validation
      of user args by `validate_style`.
    - calls `get_layout` to get a BaseLayout child class with Node
      coordinates given the tree style.
    - creates a ToyTreeMark to generate and return the tree drawing.
    - calls `get_canvas_and_axes`to get Canvas and Axes sized.
"""

from typing import Tuple, TypeVar
from loguru import logger
import numpy as np
from toyplot.canvas import Canvas
from toyplot.coordinates import Cartesian

# from toytree.annotate.src.add_scale_bar import add_axis_scale_bar_to_mark
from toytree.drawing import ToyTreeMark
from toytree.drawing.src.setup_canvas import get_canvas_and_axes
from toytree.layout import BaseLayout, LinearLayout, CircularLayout, UnrootedLayout
from toytree.style import (
    TreeStyle,
    validate_style,
    get_base_tree_style_by_name,
    tree_style_to_css_dict,
)


# from toytree.utils import ToytreeError
ToyTree = TypeVar("ToyTree")
logger = logger.bind(name="toytree")

def get_tree_style_updated_by_draw_args(tree: ToyTree, **kwargs) -> TreeStyle:
    """Return an expanded TreeStyle given user args and base treestyle.

    Parses the `Toytree.draw()` function arguments to update TreeStyle.
    - expands special 'ts' arg as `tree_style`.
    - logs warning for any unrecognized args.
    - gets a TreeStyle base.
    - validates style args and updates TreeStyle using them.
    - return the updated TreeStyle object.
    """
    # extract extra kwargs that are not in ToyTree.draw()
    extra_kwargs = kwargs.pop("kwargs")

    # remove 'ts' shortcut name and expand as a user tree_style arg
    if extra_kwargs.get("ts"):
        kwargs["tree_style"] = extra_kwargs.pop("ts")

    # warn user of any remaining (unsupported) kwargs
    if extra_kwargs:
        logger.warning(
            f"Unrecognized arguments skipped: {list(extra_kwargs)}."
            "\nCheck the docs, argument names may have changed."
        )

    # get a TreeStyle COPY from tree.style or new ts TreeStyle()
    style = get_tree_style_base(tree, tree_style=kwargs.pop('tree_style'))

    # special handling of tree_style='p' in case Ne is not a feature.
    if style.tree_style == "p" and "Ne" not in tree.features:
        style.edge_widths = None

    # check and expand user-kwargs if provided else base style value
    style = validate_style(tree, style, **kwargs)
    return style


def draw_toytree(tree: ToyTree, **kwargs) -> Tuple[Canvas, Cartesian, ToyTreeMark]:
    """Parse arguments to draw function and return drawing objects.

    The drawing style arguments can be entered in two ways, either
    by modifying attributes of the `.style` dict-like object linked
    to a ToyTree, or by providing parameter arguments to the `.draw`
    function. The latter type overrides any values in the `.style`
    settings, and similarly, if a `tree_style` argument is used this
    overrides any/all settings in `.style` setting a new fresh base
    style setting.

    This function parses arguments to `draw`, updates a copy of the
    `.style`, and passes these style args as a dict to the ToyTreeMark
    class init function, along with the edges and verts.
    """
    # extract Cartesian axes or None
    axes = kwargs.pop("axes")
    label = kwargs.pop("label")

    # get tree style expanded for user args and base tree style (ts)
    style = get_tree_style_updated_by_draw_args(tree, **kwargs)

    # get a Layout with coordinates projected based on style
    layout = get_layout(
        tree=tree,
        style=style,
        fixed_order=kwargs.pop("fixed_order"),
        fixed_position=kwargs.pop("fixed_position"),
    )

    # set tip angles on 'c' or unrooted layouts
    if style.tip_labels_angles is None:
        style.tip_labels_angles = layout.angles

    # generate toyplot Mark. Style is already validated. tables of int idx labels
    mark = ToyTreeMark(
        ntable=layout.coords,
        ttable=layout.tcoords,
        etable=tree.get_edges('idx'),
        **tree_style_to_css_dict(style),
    )

    # create Canvas and Cartesian if they don't yet exist.
    canvas, axes = get_canvas_and_axes(
        axes,
        mark,
        kwargs.get('width'),
        kwargs.get('height')
    )

    # make range symmetric for circular trees
    if style.layout == "c":
        axes.aspect = "fit-range"

    # add ToyTreeMark to Cartesian axes.
    axes.add_mark(mark)

    # Show axes with a scale bar if requested.
    if mark.scale_bar is False:
        # hide axes if Cartesian is new and scale bar not added.
        if canvas is not None:
            axes.x.show = False
            axes.y.show = False
    else:
        tree.annotate.add_axes_scale_bar(axes)

    # add label text to axes (user can add more styling outside)
    if label:
        axes.label.text = label

    return canvas, axes, mark


def get_tree_style_base(tree: ToyTree, **kwargs) -> TreeStyle:
    """Return a style class object updated by user-args.

    If a `tree_style` arg was entered this overrides the default
    base style. Any additional style options in kwargs are applied
    on top of the base style.
    """
    # allow tree_style or ts as args.
    if kwargs.get('tree_style'):
        tree_style = kwargs['tree_style']
    else:
        tree_style = kwargs.get('ts')

    # get new TreeStyle from user-entered arg (e.g., 's')
    if tree_style is not None:
        style = get_base_tree_style_by_name(tree_style[0])

    # get new TreeStyle from tree.style.tree_style (e.g., 'n')
    elif tree.style.tree_style is not None:
        style = get_base_tree_style_by_name(tree.style.tree_style[0])

    # get deepcopy of current TreeStyle in tree.style
    else:
        style = tree.style.copy()
    return style


def get_layout(tree: ToyTree, style: TreeStyle, **kwargs) -> BaseLayout:
    """Return a Layout class object given a tree, style, and optional
    positional constraints.
    """
    # return a linear layout
    if style.layout in "rlud":
        return LinearLayout(
            tree,
            style,
            kwargs.get("fixed_order"),
            kwargs.get("fixed_position"),
            kwargs.get("interior_algorithm", 0),
        )

    # warn user that positional constraints are only for linear layouts
    if kwargs.get("fixed_order") is not None:
        logger.warning(f"`fixed_order` has no effect on `{style.layout}` layout.")
    if kwargs.get("fixed_position") is not None:
        logger.warning(f"`fixed_position` has no effect on `{style.layout}` layout.")

    # return circular or unrooted layout
    if style.layout[0] == "c":
        # require tip_labels_align=True for 'c' unless no variation
        if not np.allclose(tree.get_tip_data("height"), 0):
            style.tip_labels_align = True
        return CircularLayout(tree, style)

    # anything else is unrooted (None, etc.)
    return UnrootedLayout(tree, style)
    # if style.layout in ["unrooted", "unroot", "u1", "u2"]:
    # raise ToytreeError(f"layout style not recognized: {style.layout}")


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("DEBUG")
    tre = toytree.rtree.imbtree(10, )
    tre._draw_browser('s', admixture_edges=[((2, 3), 4)], tmpdir="~", label="HI")

    # tre[0].name = "HELLO"
    # tre[4].name = "HELLO"
    # tre[6].name = "HELLO WORLD"
    # tre[-1]._dist = 10
    # tre._draw_browser(
    #     ts='p',
    #     layout='c',
    #     edge_type='c',
    #     height=400,
    #     width=400,
    #     node_mask=tre.get_node_mask(1, 5, 9),
    #     # tip_labels="name",
    # )



#!/usr/bin/env python

"""Phylogenetic ANOVA

TODO
----
This module is intended for the development of phylogenetic generalized
least squares (PGLS) methods, including a phylogenetic ANOVA. Some
pseudo-code is written but this is IN DEVELOPMENT, and not yet 
intended for use.

References
-----------
__Adams, D. C., & Collyer, M. L. (2018). Phylogenetic ANOVA: group-clade 
aggregation, biological challenges, and a refined permutation procedure.
Evolution, 72(6), 1204-1215.__

"""

from typing import Sequence
import numpy as np
from toytree.core import ToyTree


def anova(values: Sequence[float], groups: Sequence[str]) -> dict[str, float]:
    """Perform a one-way analysis of variance (ANOVA).

    ANOVA is typically used to test for differences in means across
    different groups or treatments. It compares the variance within
    groups to the variance between groups to determine if group means
    are significantly different. This method calls the scipy.stats 
    function `f_oneway` to run a one-way ANOVA.

    Parameters
    ----------
    values: Sequence[float]
        A Sequence (e.g., list, np.ndarray, pd.Series) of quantitative
        values to be analyzed.
    groups: Sequence[str]
        A Sequence (e.g., list, np.ndarray, pd.Series) of categorical
        group assignments for each value.

    Example
    -------
    >>> rng = np.random.default_rng(124)
    >>> values = rng.normal([1, 2, 3], [2, 2, 2], size=(100, 3)).flatten("F")
    >>> groups = np.repeat([1, 2, 3], 100)
    >>> anova(values, groups)
    >>> # {'F-statistic': 33.348302, 'p-value': 8.601108e-14}
    """
    from scipy.stats import f_oneway
    values = np.array(values)
    groups = np.array(groups)
    keys = set(groups)
    grouped_data = [values[groups == i] for i in keys]
    f_stat, p_value = f_oneway(*grouped_data)
    return {"F-statistic": f_stat, "p-value": p_value}


def phylogenetic_anova(trait: Sequence[float], group: Sequence[str], tree: ToyTree) -> dict[str, float]:
    """Generalized Least Squares...

    ANOVA is typically used to test for differences in means across
    different groups or treatments. It compares the variance within
    groups to the variance between groups to determine if group means
    are significantly different. When the groups are related by a
    phylogeny, it is necessary to control for the fact that some are
    expected to be more similar to each other than others. This is
    the goal of a phylogenetic ANOVA. 

    This method calls the scipy.stats 
    function `f_oneway` to run a one-way ANOVA.

    "How is the variation in a trait between different species 
    explained by their evolutionary history?"

    Look at that statistical significance of differences between groups.
    The null hypothesis is that there are no differences; the alternative
    hypothesis is that one or more groups differs from the others even
    after accounting for phylogeny. The key outputs are an F-statistic
    and p-value. The F-statistic measures variation _between-groups_ 
    relative to variation _within-groups_. Larger values indicates
    larger relative differences between groups. The p-value tests the
    null hypothesis that there is no difference.


    """
    # check that all 'group' categories are represented in the tree
    assert set(tree.get_tip_labels()) == set(group), "Mismatch between tree and 'groups' (species)"

    # get phylogenetic variance-covariance matrix
    # raise warning if tree is not ultrametic?


def phylogenetic_generalized_least_squares(values: Sequence[float], groups: Sequence[float], tree: ToyTree) -> dict[str, float]:
    """Returns a ... result of a phylogenetic generalized least squares
    analysis.

    Generalized least squares is used to model data where the errors
    are not independent and identically distributed (i.i.d.). GLS 
    allows for the modeling of heteroscedasticity (unequal variance) 
    and autocorrelation (dependence between error terms). When the
    values being modeled evolved on a phylogeny the variance-covariance
    matrix can be used to model the error structure.

    A GLS model accounts for phylogenetic non-independence by 
    incorporating the variance-covariance structure from the tree.

    DevNotes
    --------
    This is an implementation in progress where we are trying to 
    implement PGLS using only scipy and not statsmodels.

    Example
    -------
    >>> tree = toytree.rtree.unittree(ntips=10)
    >>> values = tree.pcm.simulate_continuous_brownian(1.0, tips_only=True)
    >>> groups = tree.get_tip_labels()
    >>> phylogenetic_generalized_least_squares(values, groups, tree)
    >>> # 
    """
    from scipy.linalg import inv
    from pandas import get_dummies    

    # The response variable (e.g., trait value)
    y = np.array(values)

    # Group data
    groups = np.array(groups)
    keys = set(groups)

    # check that tree tips match groups
    assert set(tree.get_tip_labels()) == keys, "Mismatch between tree and 'groups' (species)"

    # The design matrix (e.g., intercept + group dummy variables)
    X = get_dummies(groups, drop_first=True).values

    # Covariance matrix (phylogenetic covariance)
    Sigma = tree.pcm.get_corr_matrix_from_tree()

    # Compute the inverse of the covariance matrix
    Sigma_inv = inv(Sigma)

    # Calculate (X^T Σ^(-1) X)
    XT_Sigma_inv_X = X.T @ Sigma_inv @ X

    # Calculate (X^T Σ^(-1) y)
    XT_Sigma_inv_y = X.T @ Sigma_inv @ y

    # GLS solution for beta (coefficients)
    # These represent the estimated effect of each predictor (species), accounting for phylogeny.
    beta = np.linalg.inv(XT_Sigma_inv_X) @ XT_Sigma_inv_y
    print("GLS Coefficients (Beta):\n", beta)

    # Predicted values
    y_pred = X @ beta

    # Residuals
    residuals = y - y_pred

    # Weighted residual sum of squares
    # Residual Sum of Squares is how well the model fits the data after incorporating the covariance structure.
    rss = residuals.T @ Sigma_inv @ residuals
    print("Weighted Residual Sum of Squares:\n", rss)

    # Variance-covariance matrix of coefficients
    # Used to compute confidence intervals and p-values for the coefficients.
    beta_var = np.linalg.inv(XT_Sigma_inv_X)
    print("Variance-Covariance Matrix of Beta:\n", beta_var)

    # Standard errors of coefficients
    # Quantify uncertainty around the estimated coefficients
    beta_se = np.sqrt(np.diag(beta_var))
    print("Standard Errors of Coefficients:\n", beta_se)

    # To calculate p-values for the coefficients using a t-test:
    from scipy.stats import t

    # Degrees of freedom
    n, p = X.shape
    df = n - p

    # t-values
    t_values = beta / beta_se

    # Two-tailed p-values
    p_values = 2 * t.sf(np.abs(t_values), df)
    print("t-values:\n", t_values)
    print("p-values:\n", p_values)



def test_R_gls():
    """Compare results to a pgls in R ...
    """





if __name__ == "__main__":

    import toytree

    # standard ANOVA
    rng = np.random.default_rng(124)
    values = rng.normal([1, 2, 3], [2, 2, 2], size=(100, 3)).flatten("F")
    groups = np.repeat([1, 2, 3], 100)
    result = anova(values, groups)
    # print(result)

    # phylogenetic ANOVA    


    # phylogenetic GLS
    tree = toytree.rtree.unittree(6, treeheight=2, seed=123)
    tree.treenode.draw_ascii()
    traits = tree.pcm.simulate_continuous_brownian(1.0, tips_only=True)
    print(traits)
    groups = tree.get_tip_labels()
    # vcv = tree.pcm.get_vcv_matrix_from_tree(df=True)
    phylogenetic_generalized_least_squares(traits, groups, tree)



#!/usr/bin/env python

"""Function to return a mapping {Node: feature}.

This function is pretty fast while also allowing for regex matching
in Node queries. It is used in many places.
"""

from typing import TypeVar, Mapping, Any
import numpy as np
from toytree import Node, ToyTree
from toytree.utils.src.exceptions import NODE_NOT_IN_TREE_ERROR


Query = TypeVar("Query", Node, int, str)


def expand_node_mapping(tree: ToyTree, mapping: Mapping[Query, Any]) -> Mapping[Node, Any]:
    """Return a mapping of {Nodes: feature} by expanding Node queries.

    This function is similar to `get_nodes`, but in addition to
    expanding a query into Nodes it also keep their mapping to
    features. This function is used inside other toytree functions
    such as `set_node_data`.

    Parameters
    ----------
    mapping: Mapping[Query, Any]
        A mapping (e.g., dict, pd.Series) of Node queries to feature
        values. Queries can be Node objects, idx labels, str names, or
        regular expressions matching mulitple string names, indicated
        by using a '~' str prefix.

    Notes
    -----
    You can `practice your regex here: <https://pythex.org/>`_
    The order in which Nodes are returned may be different than the
    order of queries.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(8, seed=123)
    >>> mapping = {0: 1, 'r1': 2, '~r[2-3]+': 4)
    >>> tree.set_node_data('x', mapping, inplace=True)
    >>> tree.get_node_data('x')
    >>> 0     0.0
    >>> 1     0.0
    >>> 2     0.0
    >>> 3     0.0
    >>> 4     1.0
    >>> 5     2.0
    >>> 6     NaN
    >>> 7     NaN
    >>> 8     NaN
    >>> 9     NaN
    >>> 10    NaN
    >>> 11    NaN
    >>> 12    NaN
    >>> 13    NaN
    >>> 14    NaN
    """
    nodes = {}
    names = {}
    for key, value in dict(mapping).items():
        if isinstance(key, int):
            nodes[tree[key]] = value
        elif isinstance(key, Node):
            if key not in tree:
                raise ValueError(NODE_NOT_IN_TREE_ERROR)
            nodes[key] = value
        elif isinstance(key, str):
            names[key] = value  # names are expanded to Nodes below.
        elif isinstance(key, np.integer):
            nodes[tree[key]] = value
        else:
            raise TypeError(f"query type {type(key)} not supported.")

    # match Node names as a group so we only need to perform one
    # tree traversal. Each query can return multiple regex hits.
    for name, feature in names.items():
        matched = set(tree._iter_nodes_by_name_match(name))
        for node in matched:
            nodes[node] = feature
    # return expanded mapping
    return nodes


if __name__ == "__main__":

    import toytree
    tree = toytree.rtree.unittree(8)
    mapping = {"~r[0-3]+": 0, "r4": 1, tree[5]: 2}
    emap = expand_node_mapping(tree, mapping)
    print(emap)
    # tree.set_node_data('x', emap, inplace=True)
    # print(tree.get_node_data('x'))



'''
functions for determining the Blomberg's K phylogenetic signal metric

References
---------
Blomberg, Simon P., Theodore Garland, and Anthony R. Ives. 
“TESTING FOR PHYLOGENETIC SIGNAL IN COMPARATIVE DATA: BEHAVIORAL TRAITS ARE MORE LABILE.” 
Evolution 57, no. 4 (April 2003): 717–45. https://doi.org/10.1111/j.0014-3820.2003.tb00285.x.


'''
import numpy as np
import toytree





def compute_D(V):
    """
    Computes the transformation matrix D such that D V D' = I.
    """
    # Compute the inverse square root of V using eigendecomposition
    eigvals, eigvecs = np.linalg.eigh(V)  # Eigen decomposition of V
    Lambda_sqrt_inv = np.diag(1.0 / np.sqrt(eigvals))  # Inverse sqrt of eigenvalues
    D = Lambda_sqrt_inv @ eigvecs.T  # Compute D

    return D

def generalized_least_squares(X, V):
    """
    Computes the phylogenetic mean (a_hat) and residual variance (MSE)
    using Generalized Least Squares (GLS).
    """
    n = len(X)
    D = compute_D(V)  # Compute the transformation matrix D

    # Step 2: Transform X
    U = D @ X

    # Step 3: Compute phylogenetic mean
    a_hat = np.mean(U)

    # Step 4: Compute residual variance (MSE)
    residuals = U - a_hat
    MSE = (residuals.T @ residuals) / n

    return a_hat, MSE

def calculate_blomberg_k(tree, data):
    V = tree.pcm.get_vcv_matrix_from_tree()
    n = tree.ntips

    # Compute GLS phylogenetic mean and MSE
    a_hat, MSE = generalized_least_squares(data, V)

    # Compute observed MSE_0
    residuals_0 = data - a_hat
    MSE_0 = (residuals_0.T @ residuals_0) / (n - 1)

    # Compute expected MSE under Brownian motion
    V_inv = np.linalg.inv(V)
    expected_MSE = (1 / (n - 1)) * ((np.trace(V) - n / (np.sum(V_inv))))

    # Compute Blomberg's K
    K = (MSE_0 / MSE) / expected_MSE
    return K

if __name__ == "__main__":

    tree = toytree.rtree.rtree(10)
    #purposefully high signal
    tip_data_1 = np.array([3.45, 6.78, 7.32, 16.43, 16.94, 25.56, 35.66, 37.55, 67.43, 68.78]) 
    mu, sigma = 0, 0.1 
    #random data (should be low signal)
    tip_data_2 = np.random.normal(mu, sigma, 10) 

    print(calculate_blomberg_k(tree, tip_data_1))
    print(calculate_blomberg_k(tree, tip_data_2))


#!/usr/bin/env python

"""Logger for development and user warnings.

All toytree modules that use logging use a 'bound' logger that will
be filtered here to only show for toytree and not for other Python
packages.
"""

import sys
from contextlib import contextmanager
from loguru import logger
import toytree
# from toytree.utils import ToytreeError


LOGGERS = [0]


def colorize():
    """colorize the logger if stderr is IPython/Jupyter or a terminal (TTY)"""
    try:
        import IPython
        tty1 = bool(IPython.get_ipython())
    except ImportError:
        tty1 = False
    tty2 = sys.stderr.isatty()
    if tty1 or tty2:
        return True
    return False


def set_log_level(log_level="INFO"):
    """Set the log level for loguru logger bound to toytree.

    This removes default loguru handler, but leaves any others in place,
    and adds a new one that will filter to only print logs from
    toytree modules, which should use `logger.bind(name='toytree')`.

    Examples
    --------
    >>> # suppress toytree logs below INFO level
    >>> toytree.set_log_level("INFO")
    >>>
    >>> # write a log message from the toytree logger
    >>> from loguru import logger
    >>> logger.bind(name="toytree").info("logged message from toytree")
    """
    for idx in LOGGERS:
        try:
            logger.remove(idx)
        except ValueError:
            pass
    idx = logger.add(
        sink=sys.stderr,
        level=log_level,
        colorize=colorize(),
        format="{level.icon} toytree | {module}:{function} | {message}",
        filter=lambda x: x['extra'].get("name") == "toytree",
    )
    LOGGERS.append(idx)
    logger.enable("toytree")
    logger.bind(name="toytree").debug(
        f"toytree v.{toytree.__version__} logging enabled"
    )


@contextmanager
def capture_logs(level="INFO", format="{level}:{name}:{message}"):
    """Capture loguru-based logs (used in unittests mainly.)"""
    output = []
    handler_id = logger.add(output.append, level=level, format=format)
    yield output
    logger.remove(handler_id)


if __name__ == "__main__":

    toytree.set_log_level("DEBUG")
    logger.bind(name="toytree").info("THIS IS A TEST.")

    with capture_logs("INFO") as cap:
        logger.bind(name="toytree").debug("Hello")
        logger.bind(name="toytree").info("Hello2")
    print(f"Captured: {cap}")

    # logger.error("HO")



#!/usr/bin/env python

"""Relative evolutionary divergence

"""

from toytree import ToyTree
import numpy as np


__all__ = ["get_relative_evolutionary_divergence"]


def get_relative_evolutionary_divergence(tree: ToyTree, inplace: bool = False):
    """Returns a dict mapping Node idx to relative evolutionary divergence.
    
    RED interpolates the relative divergence of each node in a rooted tree by
    the average distance to the node's leafs, such that the root has a RED of
    0 and the leafs have a RED of 1.
    
    Parameters
    ----------
    tree: ToyTree
        A tree with edge lengths on which RED values will be computed.
    inplace: bool
        If True a tree is returned with data stored to Nodes as 
        feature "RED". Else a dict is returned.

    References
    ----------
    https://doi.org/10.1038/s41564-021-00918-8
    https://doi.org/10.48550/arXiv.1308.6333
    """
    # make sure the tree is rooted
    if not tree.is_rooted() :
        raise Exception("RED requires a rooted tree.")

    # store root to 0, tips to 1
    red = {tree.treenode.idx: 0}
    red.update({i: 1 for i in range(tree.ntips)})
    
    # get all-by-all node distance matrix
    mat = tree.distance.get_node_distance_matrix(topology_only=False)
    
    # preorder (parent then child) excluding root and tips
    for node in tree[-2: tree.ntips: -1]:
        
        # get parent's red value (P), and dist to parent (a)
        P = red[node.up.idx]
        a = mat[node.idx, node.up.idx]
        
        # get avg dist from this node to each of its leaves (b)
        b = np.mean([mat[node.idx, leaf.idx] for leaf in node.iter_leaves()])
        
        # integrety check in case of invalid branch lengths
        if a + b == 0:
            raise ValueError(f"node {node}: a == b == {a}")
        
        # store this nodes red value
        red[node.idx] = P + (a / (a + b)) * (1 - P)

    # return a tree with data set to nodes, or return the data in a dict
    if inplace:
        return tree.set_node_data("RED", red)
    else:
        return red


if __name__ == "__main__":

    import toytree
    tree = toytree.rtree.unittree(ntips=10, seed=123)
    # reds = tree.pcm.get_relative_evolutionary_divergence()
    reds = toytree.pcm.get_relative_evolutionary_divergence(tree)
    print(reds)


#!/usr/bin/env python

"""Topology manipulation functions for ToyTrees.

This module includes function for manipulating ToyTrees by adding,
removing, or changing the relationships among TreeNodes. It is the
preferred way for users to modify tree topologies, compared to
editing TreeNodes directly, since it ensures that the TreeNodes
have a valid ToyTree coordinate structure when returned.

Most of these functions use a Query selector...

Methods
-------
Return ToyTree with some internal nodes collapsed.
>>> collapse_nodes(tree, ...)

Return ToyTree with one Node rotated (._children tuple order reversed).
>>> rotate_node(tree, ...)

Return a ToyTree connecting a subset of Nodes in a tree.
>>> prune(tree, ...)

Return a ToyTree with some tip Nodes removed.
>>> drop_tips(tree, ...)

Return ToyTree with one or more polytomies randomly resolved.
>>> resolve_polytomies(tree, ...)

Add an internal node by splitting an edge to create new node.
>>> add_internal_node(tree, ...)

Add a tip node as a child of an existing Node.
>>> add_child_node()

Add a sister node as a child of the same existing parent Node (synonymous w/ add_child() called on parent)
>>> add_sister_node(tree, ...)

Add a parent-child pair to split an existing branch into two children (new child Node, old child clade)
>>> add_internal_node_and_child()

Add a parent-child clade pair to split an existing branch.
>>> add_internal_node_and_subtree()

"""

from typing import Optional, TypeVar, Tuple, Callable, Union
from loguru import logger
import numpy as np
from toytree.core.apis import TreeModAPI, add_subpackage_method, add_toytree_method
from toytree.core.node import Node
from toytree.core.tree import ToyTree
from toytree.utils import ToytreeError

logger = logger.bind(name="toytree")
Query = TypeVar("Query", str, int, Node)

__all__ = [
    "ladderize",
    "collapse_nodes",
    "remove_unary_nodes",
    "rotate_node",
    "extract_subtree",
    "prune",
    "drop_tips",
    "bisect",
    "resolve_polytomies",
    "add_internal_node",
    "add_child_node",
    "add_sister_node",
    "add_internal_node_and_child",
    "add_internal_node_and_subtree",
    "remove_nodes",
    "merge_nodes",
]


@add_toytree_method(ToyTree)
@add_subpackage_method(TreeModAPI)
def ladderize(tree: ToyTree, direction: bool = False, inplace: bool = False) -> ToyTree:
    """Return a ladderized tree (ordered descendants)

    In a ladderized tree nodes are rotated so that the left/
    right child always has fewer/more descendants.

    Parameters
    ----------
    direction: bool
        If False then child Nodes are sorted (left to right) from
        smallest to largest number of descendants. If True they are
        sorted in the reverse order.
    """
    # get a copy of the tree to modify
    nself = tree if inplace else tree.copy()
    direction = bool(direction)  # needed to ensure invert below against int arg

    # visit all nodes from tips to root recording size on the way
    sizes = {}
    for nidx in range(tree.nnodes):
        node = nself[nidx]

        # get node size
        if node.is_leaf():
            sizes[node.idx] = 1
        else:
            sizes[node.idx] = sum(sizes[child.idx] for child in node.children)
            node._children = tuple(sorted(
                node._children,
                key=lambda x: sizes[x.idx],
                reverse=direction,
            ))

    # update idx labels for new tree ladderization
    nself._update()
    return nself


@add_subpackage_method(TreeModAPI)
def collapse_nodes(
    tree: ToyTree,
    *query: Query,
    min_dist: float = 1e-6,
    min_support: float = 0,
    inplace: bool = False,
) -> ToyTree:
    """Return ToyTree with some internal nodes collapsed.

    Nodes can be entered as Node instances, Node names strings,
    or Node int idx labels, and/or Nodes can be selected by
    minimum dist or support values. Selected Nodes are collapsed
    into multi-furcating polytomies. For example, set
    min_support=50 to collapse all nodes with support < 50, and/or
    select Node idx 10 to collapse Node 10.

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. To select internal Nodes use idx labels
        or Node object from `tree.get_mrca_node('name1', 'name2')`.
    min_dist: float
        The minimum dist (edge length) value allowed.
    min_support: float
        The minimum support (e.g., bootstrap) value allowed.
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    Note
    ----
    This cannot be used to remove the root Node. To collapse the root
    into a polytomy use `toytree.mod.unroot`, or collapse the Node(s)
    directly below the root Node.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=12)
    >>> tree.mod.collapse_nodes(14)
    >>> tree.mod.collapse_nodes(tree.get_mrca_node('r1', 'r2').idx)
    >>> tree.mod.collapse_nodes(tree.get_mrca_node('~r[1-4]').idx)
    >>> tree = tree.set_node_data("dist", {22: 0.005, 21: 0.005})
    >>> tree = tree.set_node_data("support", {20: 50}, default=100)
    >>> tree = tree.mod.collapse_nodes(min_dist=0.01, min_support=45)
    """
    nodes = () if not query else tree.get_nodes(*query)
    if not inplace:
        tree = tree.copy()
        nodes = [tree[i.idx] for i in nodes]

    # iterate over all internal nodes
    for node in tree[tree.ntips:-1]:
        if (node.dist < min_dist) | (node.support < min_support) | (node in nodes):
            node._delete(preserve_dists=True, prevent_unary=True)
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def merge_nodes(
    tree: ToyTree,
    merge_method: Union[Callable, str] = "name",
    selection_method: Callable = min,
    inplace: bool = False,
) -> ToyTree:
    r"""Return ToyTree with one or more Nodes merged based on a Callable.

    Merging means to discard at least one tip and one internal Node
    while keeping one child node that inherits its parents dists. An
    example use case for this is to compress duplicate tips in a clade
    into a single representative.

                    |                 |
                   2()                |
                   /  \     --->      |
                0(a)  1(a)           0(a)

    Parameters
    ----------
    merge_method: Callable or str
        Two options: (1) A function that returns True if a Node should
        be merged; or (2) A feature name for which a Node will be
        merged if all descendant leaves share the same feature value.
    selection_method: Callable
        A function that returns a single Node from a collection of
        Nodes. The default func `min` will return the Node with
        the lowest idx index.
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    See Also
    --------
    `toytree.mod.remove_nodes`

    Example
    -------
    >>> tree = toytree.rtree.unittree(5, seed=123)
    >>> tree1 = tree.mod.add_internal_node_and_child("r1", name="r1")
    >>> # merge nodes with identical leaf names.
    >>> tree2 = tree1.mod.merge_nodes("name")
    >>> # more verbose example to do the same
    >>> merge_method = lambda x: len(set(x.iter_leaf_names())) == 1
    >>> tree2 = tree1.mod.merge_nodes(merge_method)
    >>> toytree.mtree([tree1, tree2]).draw();
    """
    # select which nodes will be removed
    merge = set()

    # if feature selected make a def to return True if all leaf values
    # are the same.
    if isinstance(merge_method, str):
        feat = merge_method
        def merge_method(node: Node) -> bool:
            nvals = set(getattr(i, feat) for i in node.iter_leaves())
            return len(nvals) == 1

    # check Callables
    assert isinstance(merge_method(tree[0]), bool), "merge_method should return a boolean"
    assert len(selection_method(tree[:3])) == 1, "selection_method should retain only one Node"

    # iterate over tree from root to tips
    for node in tree[::-1][1:]:
        if node not in merge:
            if merge_method(node):
                desc = set(node.iter_descendants())
                keep = selection_method(desc)
                desc.discard(keep)
                merge.update(desc)

    # remove nodes and return tree
    return tree.mod.remove_nodes(*merge, inplace=inplace)


@add_subpackage_method(TreeModAPI)
def remove_nodes(tree: ToyTree, *query: Query, preserve_dists: bool = True, inplace: bool = False) -> ToyTree:
    """Return ToyTree with one or more Nodes removed.

    If multiple Nodes are entered they are removed in a postorder
    traversal of the tree. Nodes can be selected using Node Queries.

    Parameters
    ----------
    *query: str, int, or Node
        The Node to rotate can be selected by entering the Node object,
        or its idx label, or name str. For internal Nodes, multiple
        queries can be entered and their MRCA will be rotated.
    preserve_dists: bool
        If True then children inherit the dist values of their
        deleted parents so that their dist value reaches their
        grandparents original height. If False, children retain
        their original dist but are connected to their grandparent.
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    See Also
    --------
    `toytree.mod.remove_unary_nodes`

    Example
    -------
    >>> # remove a tip leaving behind a unary internal Node.
    >>> tree = toytree.rtree.unittree(5, seed=123)
    >>> tree = toytree.mod.remove_node(tree, "r0")
    """
    nodes = () if not query else tree.get_nodes(*query)
    if not inplace:
        tree = tree.copy()
        nodes = [tree[i.idx] for i in nodes]

    # postorder idx traversal
    for node in tree:
        if node in nodes:
            node._delete(preserve_dists, prevent_unary=False)
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def remove_unary_nodes(tree: ToyTree, inplace: bool = False) -> ToyTree:
    r"""Return ToyTree with any unary Nodes removed.

                4     remove_unary_nodes     4
               / \                          / \
              3   2         ------>        /   \
             /     \                      /     \
            0       1                    0       1

    Parameters
    ----------
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    Example
    -------
    >>> # add two unary Nodes then remove them.
    >>> tree = toytree.rtree.unittree(5, seed=123)
    >>> tree = toytree.mod.add_internal_node(tree, "~r[0-2]", name="i")
    >>> tree = toytree.mod.add_internal_node(tree, "~r[0-1]", name="j")
    >>> tree = toytree.mod.remove_unary_nodes(tree)
    """
    tree = tree if inplace else tree.copy()
    tipset = set(tree[i] for i in range(tree.ntips))
    for node in tree.traverse("postorder"):
        if len(node.children) == 2:
            tipset.add(node)
        if node not in tipset:
            new_parent = node._up
            if new_parent:
                true_node = node._children[0]
                new_parent._add_child(true_node)
                new_parent._remove_child(node)
                true_node._dist += node._dist
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def rotate_node(tree: ToyTree, *query: Query, inplace: bool = False) -> ToyTree:
    """Return ToyTree with one Node rotated (children order reversed).

    Rotates *only one Node per call*. Internal Nodes can be selected
    by idx label, or by entering multiple tip Node names from which the
    MRCA will be selected and rotated.

    Parameters
    ----------
    *query: str, int, or Node
        The Node to rotate can be selected by entering the Node object,
        or its idx label, or name str. For internal Nodes, multiple
        queries can be entered and their MRCA will be rotated.
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10)
    >>> toytree.mod.rotate_node(tree, 12)
    >>> toytree.mod.rotate_node(tree, 'r0', 'r1')
    >>> toytree.mod.rotate_node(tree, '~r[0-3]$')
    >>> # Can chain multiple calls together:
    >>> tree.mod.rotate_node(14).mod.rotate_node(13).mod.rotate_node(12)
    """
    node = tree.get_mrca_node(*query)
    if not inplace:
        tree = tree.copy()
        node = tree[node.idx]
    node._children = tuple(node.children[::-1])
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def extract_subtree(tree: ToyTree, *query: Query) -> ToyTree:
    r"""Return a subtree/clade extracted from a larger tree as a ToyTree.

                4      extract_subtree (3)
               / \                           |
              3   \         ------>          3
             / \   \                        / \
            0   1   2                      0   1

    Note
    ----
    The returned subtree is a copy of the subtree from the original;
    the original tree is not affected.

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors (Node object, names, or idx labels)
        from which the MRCA is found. This will serve as the root Node
        of the returned tree.

    See Also
    --------
    `toytree.mod.bisect`
    `toytree.mod.prune`
    `toytree.mod.drop_tips`

    Example
    -------
    >>> tree = toytree.tree("((a,b),c);")
    >>> subtree = tree.mod.extract_subtree('a', b')
    >>> toytree.mtree([tree, subtree]).draw('p')
    """
    node = tree.get_mrca_node(*query)
    return ToyTree(node.copy(detach=True))


@add_subpackage_method(TreeModAPI)
def bisect(tree: ToyTree, *query: Query, reroot: bool = False, dist_partition: float = 1.0) -> Tuple[ToyTree, ToyTree]:
    r"""Return a tree bisected into two subtrees on a selected edge.

    This returns two bisected subsets of the original tree. It does not
    affect the original tree object. If query selects the treenode on
    a rooted tree it returns a subtree for each child as a treenode with
    its original dist. If query selects the treenode on an urooted tree
    it raises an error. If query selects any other node it will split
    the edge above to create two subtrees. The query node will become
    a treenode of one new subtree. The arg 'reroot=False` will retain
    the original treenode in the other subtree, otherwise the node above
    the query becomes the new treenode (i.e., the subtree is re-rooted).
    The subtree below the query will inherit all of the dist of the
    split edge unless the dist_partition arg is used to designate the
    proportion of dist (0.0-1.0) to assign to this treenode.

           __6(T)__      bisect (6)
          |       |       ------->          |          |
         _4_     _5_                       4(T)       5(T)
        |   |   |   |                     |   |      |   |
        0   1   2   3                     0   1      2   3
    # bisect on treenode (T) of rooted tree returns the two child trees.

           __6(T)__      bisect (4)         |
          |       |       ------->          |
         _4_     _5_                       4(T)       5(T)
        |   |   |   |                     |   |      |   |
        0   1   2   3                     0   1      2   3
    # bisect on internal node (4) returns tree rooted on query (4) with
    # full dist of split edge. The treenode (T) is ignored if on split
    # edge and subtree from the side of split edge (5) is returned.

           __6(T)__       bisect (4)
          |       |   dist_partition=0.5    |          |
         _4_     _5_      ------->         4(T)       5(T)
        |   |   |   |                     |   |      |   |
        0   1   2   3                     0   1      2   3
    # The dist of the split edge can be partitioned among the two
    # subtrees using `dist_partition`.

           __6(T)__       bisect (1)                4(T)
          |       |      reroot=True               |   |
         _4_     _5_      ------->                 0   |
        |   |   |   |                       |         _5_
        0   1   2   3                      1(T)      |   |
                                                     2   3
    # If `reroot=True` the subtree on the other side of the split edge
    # is rerooted to make that node the treenode.

           __6(T)__       bisect (1)                 _6(T)
          |       |      reroot=False               |    |
         _4_     _5_      ------->                  |   _5_
        |   |   |   |                       |       |  |   |
        0   1   2   3                      1(T)     0  2   3
    # If `reroot=False` the original treenode remains the treenode on
    # the second subtree and any unary nodes (4) are removed.

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors (Node object, names, or idx labels)
        from which the MRCA will select the Node below the edge to cut.
        is found. This will serve as the root Node of the returned tree.
    reroot: bool
        If True the subtree above the query will be rerooted on the
        node query parent rather than retain the original treenode.
    dist_partition: float
        The proportion of dist of the split edge to assign to the
        subtree selected by query. Default is to assign all dist to
        the treenode of this subtree. A value between 0.0-1.0 can be
        entered to split the dist and assign this proportion to the
        query treenode. This arg cannot be combined with reroot=False.

    TODO
    ----
    instead of disallowing dist_partition when not re-rooting, it could
    create a new Node on the split branch to inherit the dist if the
    dist_partition value is <1.

    Examples
    --------
    >>> tree = toytree.tree.unittree(ntips=6, seed=123)
    >>> subtree1, subtree2 = tree.bisect("r2", "r3")
    >>> subtree1, subtree2 = tree.bisect("r2", "r3", reroot=True)
    >>> subtree1, subtree2 = tree.bisect("r2", "r3", reroot=True, dist_partition=0.5)
    """
    # allow node query from outside tree
    node = tree.get_mrca_node(*query)
    tree = tree.copy()
    node = tree[node.idx]

    # if treenode selected on an unrooted tree raise error
    if node.is_root() and not tree.is_rooted():
        msg = "cannot bisect on treenode of an unrooted tree. Select one of its children."
        logger.error(msg)
        raise ToytreeError(msg)

    # if treenode selected on a rooted tree return child subtrees
    elif node.is_root() and tree.is_rooted():
        children = tree.treenode.children
        left = tree.mod.extract_subtree(children[0])
        right = tree.mod.extract_subtree(children[1])
        # optionally re-partition dist root edge dist among children
        if dist_partition:
            total_dist = left.treenode.dist + right.treenode.dist
            left.treenode._dist = total_dist * dist_partition
            right.treenode._dist = total_dist * (1. - dist_partition)
        return left, right

    # ...
    if reroot:
        raise NotImplementedError("TODO")
    # root on node; detach two child clades; assign dists
    # rtree = tree.root(node)
    # pdist = sum(i.dist for i in rtree.treenode.children)
    # left, right = rtree.treenode.children

    # cleanup original tree
    ndist = node._dist
    if node.up:
        node.up._delete()
    subtree = ToyTree(node._detach())
    subtree.treenode._dist = ndist
    tree._update()
    return (subtree, tree)
    # unroot the trees
    # if not reroot:
    #     if left.ntips > 2:
    #         left.unroot(inplace=True)
    #     if right.ntips > 2:
    #         right.unroot(inplace=True)
    # return partitioned dist on rooted trees
    # else:
    #     # force to be in [0, 1]
    #     dist_partition = max(0, min(1, dist_partition))
    #     alt_partition = 1 - dist_partition
    #     left.treenode._dist = pdist - (pdist * dist_partition)
    #     right.treenode._dist = pdist - (pdist * alt_partition)
    # return (left, right)


@add_subpackage_method(TreeModAPI)
def prune(
    tree: ToyTree,
    *query: Query,
    preserve_dists: bool = True,
    require_root: bool = False,
    inplace: bool = False,
) -> ToyTree:
    r"""Return a tree w/ the minimal edges connecting a subset of Nodes.

    All nodes not included in the entered 'nodes' list will be
    removed from the topology, and the mininal spanning edges to
    connect the remaining nodes are retained. The original root node
    is preserved if 'require_root=True', otherwise the lowest
    mrca connecting the selected nodes will be kept as the new root.

                4      prune([0,1])
               / \
              3   \      ------>        3
             / \   \                   / \
            0   1   2                 0   1
    # min spanning tree of 0,1 involves only nodes 0,1,3

                4      prune([0,1])     4
               / \    require_root=1    |
              3   \      ------>        3
             / \   \                   / \
            0   1   2                 0   1
    # min spanning tree of 0,2 does not require node 4 but it is kept.

                4      prune([0,2])      4
               / \   preserve_dists=0   / \
              3   \      ------>       0   \
             / \   \                        \
            0   1   2                        2
    # min spanning tree of 0,2 involves only 0,2,4 (node 3 is excluded).
    # if preserve_dists=False its dist is also discarded.

                4    prune([0,2,3])      4
               / \                      / \
              3   \      ------>       3   \
             / \   \                  /     \
            0   1   2                0       2
    # min spanning tree of 0,2,3 only excludes Node 1. Internal nodes
    # are kept if selected in query, even if they are unary.

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. You can select tip Nodes and/or internal
        Nodes to be kept in the tree.
    preserve_dists: bool
        If True then the edge lengths of internal nodes that are
        removed are merged into the 'dist' attribute of their
        descendant node that is preserved.
    require_root: bool
        If True then the root node is always preserved. If False
        then the root is only preserved if is it the mrca of the
        selected nodes, otherwise the mrca Node is returned.
    inplace: bool
        If True then the original tree is changed in-place, and also
        returned, else a copy is returned and original is unchanged.

    Examples
    --------
    >>> tree = toytree.tree("((a,b)ab,c)r;")
    >>> tree.mod.prune("a", "b", require_root=False)
    >>> # (a,b)ab;
    >>> tree.mod.prune("a", "b", require_root=True)
    >>> # ((a,b)ab)r;
    >>> tree.mod.prune("a", "b", "ab", require_root=False)
    >>> # (a,b)ab;
    >>> tree.mod.prune("a", "b", "ab", require_root=True)
    >>> # ((a,b)ab)r;
    """
    nodes = tree.get_nodes(*query)
    if not inplace:
        tree = tree.copy()
        nodes = [tree[i.idx] for i in nodes]

    # get mrca node if not requiring root
    if not require_root:
        mrca = tree.get_mrca_node(*nodes)
        nodes.append(mrca)

    # traverse postorder removing nodes, but keeping treenode for now
    counter = {i: 0 for i in tree}
    for node in tree[:-1]:
        # increment parent count and skip node if it is in query
        if node in nodes:
            counter[node._up] += 1
            continue

        # increment parent count if in counter, skip if >1 in counter.
        if counter[node]:
            counter[node._up] += 1
            if counter[node] > 1:
                continue

        # remove the node
        node._delete(preserve_dists=preserve_dists, prevent_unary=False)

    # set mrca node as new treenode or update and return current tree
    if not require_root:
        return ToyTree(mrca._detach())
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def drop_tips(
    tree: ToyTree,
    *query: Query,
    inplace: bool = False,
) -> ToyTree:
    """Return a ToyTree with some tip Nodes removed.

    The ToyTree with the selected tip Nodes (and any remaining internal
    nodes without children) are removed while retaining the original
    edge lengths between remaining nodes. This is effectively the
    inverse of `prune`. Tip names can be selected using a Query method
    of Node instances, Node names, or Node idx int labels. Only
    selected tip Nodes affect the result.

    Parameters
    ----------
    tree: ToyTree
        An input ToyTree to perform function on.
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels.
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    See Also
    --------
    `mod.extract_subtree`
        Extract a complete clade from a tree as a new ToyTree.
    `mod.prune`
        Extract the minimal spanning tree from a tree that connects
        a subset of tip or internal Nodes.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10)
    >>> tree.mod.drop_tips(1, 2, 3).draw()
    >>> tree.mod.drop_tips('~r[0-3]$').draw()
    >>> tree.mod.drop_tips('r1', 'r2')
    """
    # raise exception if no nodes were selected
    if not query:
        msg = "No nodes selected. Enter a node query."
        logger.error(msg)
        raise ValueError(msg)

    # get query and tree
    nodes = tree.get_nodes(*query)
    if not inplace:
        tree = tree.copy()
        nodes = [tree[i.idx] for i in nodes]

    # raise exception if all tips were selected
    if len(nodes) == tree.nnodes:
        msg = "Cannot drop all tips from the tree."
        logger.error(msg)
        raise ValueError(msg)
    internal = []
    for node in sorted(nodes):
        if node.is_leaf():
            node._delete(prevent_unary=True)
        else:
            internal.append(node)
    # warn user that internal nodes were ignored.
    if internal:
        logger.warning("Only tip Nodes are removed. See `mod.remove_nodes`.")
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def resolve_polytomies(
    tree: ToyTree,
    *query: Query,
    dist: float = 1e-6,
    support: float = np.nan,
    recursive: bool = True,
    seed: Optional[int] = None,
    inplace: bool = False,
) -> ToyTree:
    """Return ToyTree with one or more polytomies randomly resolved.

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. If no Nodes are selected then ALL nodes
        that are multifurcating will be resolved.
    dist: float
        The dist value to set on newly created nodes.
    support: float or np.nan
        The support value to set on newly created nodes.
    recursive: bool
        Recursively resolve nested polytomies (default=True); if False
        then a n-tomy will only be resolved into a (n-1)-tomy.
    seed: int or None
        A seed for numpy random generator for reproducible resolving.
    inplace: bool
        If True then the original tree is changed in-place, and
        returned, rather than leaving original tree unchanged.

    Examples
    --------
    >>> tree = toytree.tree("((a,b,c),d);")
    >>> tree.mod.resolve_polytomies().draw();
    """
    nodes = tree.get_nodes(*query)
    if not inplace:
        tree = tree.copy()
        nodes = [tree[i.idx] for i in nodes]
    rng = np.random.default_rng(seed)

    for node in nodes:
        _resolve_nodes(
            node=node, dist=dist, support=support,
            rng=rng, recursive=recursive)
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def add_internal_node(
    tree: ToyTree,
    *query: Query,
    dist: Optional[float] = None,
    name: Optional[str] = None,
    inplace: bool = False,
) -> ToyTree:
    r"""Add an internal node by splitting an edge to create new node.

    Splits a branch spanning from node (A) to its parent (B)
    to create a new internal unary node (C).

                B                      B          Example
               / \                    / \         -------
              /   \      ---->       C   \        query="A"
             /     \                /     \       name="C"
            A       X              A       X      dist=None

    See Also
    --------
    :func:`.add_child_node`:
        Add a new child Node. Useful to call after `add_internal_node`.
    :func:`.add_internal_node_and_child`:
        This combines `add_internal_node` with `add_child_node`.

    Note
    ----
    If the query selects the root Node then a new Node will be inserted
    above it which will become the new root Node.

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. If multiple are entered the MRCA node will
        be used as the base of the edge to split.
    dist: float
        The distance from the selected Node at which to insert
        the new Node. This will be set as the dist of the
        descendant (A) whereas the new Node's dist (C) will be
        the descendants prior dist minus this dist value. If
        no dist value is set then the edge midpoint is used, except
        for the special case of adding a Node above the root, which
        defaults to dist=1.0.
    name: Optional[str]
        A name string to apply to the new Node. Default="".
    inplace: bool
        Modify tree in place.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(ntips=5, seed=123)
    >>> tree = tree.mod.add_internal_node('r0', name="x")
    >>> tree.draw(ts='n', node_sizes=15, node_labels="name")
    """
    # expand query
    node = tree.get_mrca_node(*query)
    if not inplace:
        tree = tree.copy()
        node = tree[node.idx]

    # get insertion edge and dist of the new Node
    parent = node.up

    # ROOT: if no parent (b/c node is root) then simply add new parent
    if not parent:
        new_node = Node(name=name if name is not None else "", dist=1.0)
        new_node._add_child(node)
        node._dist = dist if dist is not None else 1.0
        tree.treenode = new_node
        tree._update()
        return tree

    # NORMAL: parent is not root.
    dist = dist if dist is not None else node.dist / 2.
    if not node.dist >= dist >= 0:
        msg = f"the new Node dist must be >=0 and <={node.dist:.12g} (dist of {node})"
        logger.error(msg)
        raise ValueError(msg)

    # create the new Node and mend connections nearby
    new_node = Node(
        name=name if name is not None else "",
        dist=node.dist - dist,
    )

    # connect new_node to child (node) and grandparent (parent)
    node._dist = dist
    parent._remove_child(node)
    parent._add_child(new_node)
    new_node._up = parent
    new_node._children = (node, )
    node._up = new_node
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def add_child_node(
    tree: ToyTree,
    *query: Query,
    name: Optional[str] = None,
    dist: Optional[float] = None,
    inplace: bool = False,
) -> ToyTree:
    r"""Add a child Node to existing Node on a tree.

    This function selects an existing Node to act as the parent, and
    inserts a new Node as a child of that parent. The new child's name
    and dist can be set. If no dist value is entered then the child's
    dist is set to match that of its new sister Nodes, or 1.0.

                B                      B          Example
               / \                    /|\         --------
              /   \      ---->       / | \        query="B"
             /     \                /  |  \       name="D"
            A       X              A   D   X      dist=None

    Parameters
    ----------
    *query: Query
        One or more Node queries (Node object, int idx label or str
        name) from which the MRCA is found. The new Node will be added
        as a child of the MRCA Node.
    name: str or None
        Name of the new child Node.
    dist: float or None
        Edge length (dist) of the new child Node. If None it will be
        set to the same length as its longest sister, unless no sisters
        exist, then it is set to 1.
    inplace: bool
        If True the tree is modified inplace and returned, else a copy
        is returned.

    See Also
    --------
    :func:`.add_internal_node`:
        Add a new internal Node. Useful as parent to a new child Node.
    :func:`.add_internal_node_and_child`:
        This combines `add_internal_node` with `add_child_node`.

    Example
    -------
    >>> tree = toytree.rtree.imbtree(5)
    >>> tree = tree.mod.add_child_node('r0', 'r1', 'r2', name="C")
    >>> tree.draw("p", layout="r", node_labels="name");
    """
    # expand query
    node = tree.get_mrca_node(*query)
    if not inplace:
        tree = tree.copy()
        node = tree[node.idx]

    # set dist to user value else to max sister's dist, else to 1.
    if dist is None:
        if node.children:
            dist = max(child.dist for child in node.children)
        else:
            dist = 1.0
    dist = max(0, dist)

    # create the new Node that will be a child.
    new_node = Node(name=name if name is not None else "", dist=dist)

    # add as a new child to end of parent's children
    node._add_child(new_node)
    tree._update()
    return tree


@add_subpackage_method(TreeModAPI)
def add_sister_node(
    tree: ToyTree,
    *query: Query,
    name: Optional[str] = None,
    dist: Optional[float] = None,
    inplace: bool = False,
) -> ToyTree:
    r"""Add a sister Node to existing Node on a tree.

    This function selects an existing Node to act as the sister, and
    inserts a new Node as a child of the same parent. The new sister
    Node's name and dist can be set. If no dist value is entered then
    the its dist is set to match the max dist of its sister Nodes or 1.
    This function is equivalent to :func:`.add_child_node` but the
    query is used to select a sister rather than a parent.

                B                      B          Example
               / \                    /|\         --------
              /   \      ---->       / | \        query="A"
             /     \                /  |  \       name="D"
            A       X              A   D   X      dist=None

    See Also
    --------
    :func:`.add_internal_node`:
        Add a new internal Node. Useful as parent to a new sister Node.
    :func:`.add_internal_node_and_child`:
        This combines `add_internal_node` with `add_child_node`.

    Example
    -------
    >>> tree = toytree.rtree.imbtree(5)
    >>> tree = tree.mod.add_sister_node('r0', 'r1', 'r2', name="C")
    >>> tree.draw("p", layout="r", node_labels="name");
    """
    # expand query
    node = tree.get_mrca_node(*query)
    if not inplace:
        tree = tree.copy()
        node = tree[node.idx]

    assert not node.is_root(), (
        "Cannot add sister to root, it has no parent. See `add_child_node()`.")
    # simply call add_child to the parent of the selected Node.
    parent = node._up
    return add_child_node(tree, parent, name=name, dist=dist, inplace=inplace)


@add_subpackage_method(TreeModAPI)
def add_internal_node_and_child(
    tree: ToyTree,
    *query: Query,
    name: Optional[str] = None,
    dist: Optional[float] = None,
    parent_dist: Optional[float] = None,
    parent_name: Optional[str] = None,
    inplace: bool = False,
) -> ToyTree:
    r"""Add a (parent, child) edge to split an existing branch.

    Splits a branch spanning from query node (A) to its parent (B)
    to create a new internal Node (C) and child Node (D). The new
    parent and child Nodes can be given names and dist values. If
    no value is entered for `parent_dist` then the parent Node is
    inserted at the midpoint of the edge. If a parent_dist value is
    entered then it must fit within the length of the query Node's
    dist or an error is raised. The new child Node dist is not
    constrained. If no value is entered then it will be automatically
    set to match the dist of its sister Node.

                B                      B          Example
               / \                    / \         --------
              /   \      ---->       C   \        query="A"
             /     \                / \   \       name="D"
            A       X              A   D   X      parent_name="C"

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. If multiple are entered the MRCA node will
        be used as the base of the edge to split.
    name: str
        Optional name for the new child node. Default="".
    dist: float
        The dist (edge length) of the new child Node. If None this will
        be set to match its max sister's dist, if present, else 1.
    parent_dist: float
        The dist (edge length) of the new internal parent Node. If
        None it is set to the edge midpoint. If a value is entered it
        must be >0 and < query Node dist.
    parent_name: str
        Optional name for the new internal parent node. Default="".
    inplace: bool
        If False (default) a copy of the original tree is returned.

    Example
    -------
    >>> tree = toytree.rtree.imbtree(5)
    >>> tree = tree.mod.add_internal_node_and_child(
    >>>     'r3', name="C", parent_name="P")
    >>> tree.draw(node_sizes=15, node_labels="name");
    """
    # expand query
    node = tree.get_mrca_node(*query)
    if not inplace:
        tree = tree.copy()
        node = tree[node.idx]

    # set parent name to something unique
    pname = "PARENT@@@@@"
    add_internal_node(tree, node, name=pname, dist=parent_dist, inplace=True)
    add_child_node(tree, pname, name=name, dist=dist, inplace=True)

    # set parent name to user option or ''
    tree.get_nodes(pname)[0].name = parent_name if parent_name else ""
    return tree


@add_subpackage_method(TreeModAPI)
def add_internal_node_and_subtree(
    tree: ToyTree,
    *query: Query,
    subtree: ToyTree,
    subtree_stem_dist: Optional[float] = None,
    subtree_rescale: bool = False,
    parent_dist: Optional[float] = None,
    parent_name: Optional[str] = None,
    inplace: bool = False,
) -> ToyTree:
    r"""Add a subtree by splitting an edge to create a new parent
    Node and inserting the subtree as a child (i.e., tree-grafting).

    Splits a branch spanning from node (A) to its parent (C) to
    create a new ancestral node (Z) from which a subtree (Y) will
    be inserted. The name of node (Z) and the stem dist of node (Y)
    can be set. By default, if left as None, the stem dist will be
    set to half the distance to the farthest leaf, and the subtree
    will be scaled to fill the other half distance so that it aligns
    as the farthest tip node distance.

            C                       C        Example
           / \                     Z \       -------
          /   \                   /\  \      query="A"
         /     \      ---->      /  Y  \     subtree=tree('(W,X)Y;')
        /       \               /  / \  \    parent_name="Z"
       A         B             A  W   X  B

    Parameters
    ----------
    *query: str, int, or Node
        One or more Node selectors, which can be Node objects, names,
        or int idx labels. If multiple are entered the MRCA node will
        be used as the base of the edge to split.
    subtree: ToyTree or Node
        A subtree to insert into the target tree.
    subtree_stem_dist: float or None
        Edge length of the subtree stem dist (the subtree treenode
        dist value is ignored). If None it is set to half the dist of
        its sister clade.
    subtree_rescale: bool
        If True the subtree edges are rescaled to fit in the dist
        between the sister node height and stem.
    parent_dist: float or None
        Distance above the query Node at which the parent Node will be
        inserted. This length is constrained by the dist of the query
        Node. If None the midpoint of the edge is used.
    parent_name: str
        Optional name for the new internal parent node.
    inplace: bool
        If False (default) a copy of the original tree is returned.

    See Also
    --------
    ...

    Note
    ----
    On speed...

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> subtree = tree.mod.extract_subtree("r0", "r1", "r2")
    >>> new_tree = tree.mod.add_internal_node_and_subtree(
    >>>     "r4", subtree=subtree, subtree_rescale=True)
    >>> new_tree.draw();
    """
    # expand query while allowing for Node selector as input
    node = tree.get_mrca_node(*query)
    if not inplace:
        tree = tree.copy()
        node = tree[node.idx]

    # always copy the subtree and get as a detached Node.
    if isinstance(subtree, ToyTree):
        subtree = subtree.copy()
    elif isinstance(subtree, Node):
        subtree = ToyTree(subtree.copy(detach=True))
    elif isinstance(subtree, str):
        subtree = ToyTree(subtree)
    else:
        raise TypeError("subtree arg must be a ToyTree or Node instance.")

    # use a temporary parent name and set the proper name later. This
    # performs an _update call to relabel Node idxs and heights.
    pname = "PARENT@@@@@"
    add_internal_node(tree, node, name=pname, dist=parent_dist, inplace=True)

    # get new parent Node and set its name.
    parent = tree.get_nodes(pname)[0]
    parent.name = parent_name if parent_name else ""

    # get STEM dist for the subtree parent Node
    if subtree_stem_dist is None:
        subtree_stem_dist = max(i.dist for i in parent.children) / 2.

    # optional: scale subtree edges to fit in same dist as sister.
    if subtree_rescale:
        if subtree.nnodes > 1:
            remaining_dist = parent.height - subtree_stem_dist
            subtree = subtree.mod.edges_scale_to_root_height(remaining_dist)
        else:
            subtree_stem_dist = parent.children[0]._dist

    # add as a new child to end of parent's children
    parent._add_child(subtree.treenode)

    # set the subtree stem dist
    subtree.treenode._dist = subtree_stem_dist
    tree._update()
    return tree


# NEEDS WORK, not yet exposed to API.
# JUST USE ONE OF THE TREE MOVE METHODS, right?
# def move_clade(
#     tree: ToyTree,
#     idx0: int,
#     idx1: int,
#     height: Optional[float]=None,
#     name: str="",
#     shrink: bool=False,
#     inplace: bool=False,
# ):
#     r"""Move a clade from one part of the tree to another (SPR).

#     Splits a branch spanning from node idx1 (Y) to its parent (Z)
#     to create a new ancestral node (X). The idx0 clade is detached
#     from the tree (D) and its parent is removed if it leaves only
#     one child (E). The detached clade is reattached to the new
#     ancestral node (D -> X) while maintaining the edge lengths of
#     the subtree. The length of the edge connecting idx0 to the new
#     node (D, X) is automatically set to maintain ultrametricity,
#     unless the subclade height is greater than the dist arg. This
#     will raise an error unless shrink=True, in which case the
#     subtree is scaled to half the dist height to fit.

#                           Z
#           Z              / \              Z
#          / \            E   \            / \
#         E   \          /     \          /   X --- auto-set dist
#        / \   \        C       Y        /   / \
#       D   \   \  -->            -->   /   /   D --- auto-set dist
#      / \   \   \          D          /   /   / \
#     A   B   C   Y        / \        A   Y   A   B --- maintained dist
#                         A   B

#     Note
#     ----
#     This operation is synonymous with a rooted subprune regrafting
#     (SPR) move. Internal nodes that leave no descendants (e.g., E)
#     from above are removed from the tree.

#     See Also
#     --------
#     `toytree.mod.move_spr`: Faster SPR move function, but without
#     arguments to modify edge lengths like in this function.

#     Parameters
#     ----------
#     tree: ToyTree
#         A tree object.
#     idx0: int
#         The idx label of the Node at the top of the clade that
#         will be pruned from the tree and re-attached elsewhere.
#     idx1: int
#         The idx label of the Node whose branch will be split to
#         create a new parent node of the re-attached clade.
#     height: float or None
#         The height at which to insert node idx0 above node idx1.
#         This value must be in the interval between the height of
#         idx1 and its parent (node.height, node.height + node.dist).
#         If None then the node will be inserted at the midpoint
#         along node idx1's edge.
#     name: str
#         A name string to apply to the newly created internal node.
#     shrink: bool
#         If shrink is True then the subtree edge lengths (dists)
#         will be scaled to allow the clade to be inserted anywhere
#         in the tree. If False a ToyTreeError is raised if the
#         subclade cannot be inserted below the dist value.

#     Returns
#     -------
#     ToyTree
#         A modified copy of the original tree is returned.
#     """
#     # TODO: check tree_move SPR code and use that?
#     # create a copy
#     tree = tree if inplace else tree.copy()

#     # get selected nodes (FIXME: use nas to allow names selections)
#     src = tree[idx0]
#     dest = tree[idx1]

#     # cannot move a clade to its own children in rooted move
#     assert idx0 != idx1, "idx0 must be different from idx1"
#     assert idx0 != tree.treenode, "src node cannot be root."
#     assert idx1 != tree.treenode, "dest node cannot be root."
#     assert dest not in src.get_descendants(), (
#         "dest node cannot be a descendant of src in a rooted SPR move."
#     )

#     # check the height arg
#     if height is not None:
#         assert dest.up.height >= height >= dest.height, (
#             f"height {height} does occur on the node {idx1}'s edge: "
#             f"({dest.height}-{dest.up.height})")

#     # get dist arg from height
#     if height is None:
#         dist = dest.dist / 2.
#     else:
#         dist = height - dest.height

#     # if the dest is the src's parent, then we only need to modify the
#     # dist of dest and return the tree.
#     if dest == src.up:
#         diff = dest.dist - dist
#         dest.dist = dist
#         for child in dest.children:
#             child.dist += diff
#         return tree

#     # create new internal node.
#     newnode = Node(name=str(name), dist=dest.dist - dist)

#     # detach the source clade.
#     ancestor = src.up
#     src = src.detach()

#     # remove the null node (preserving bls) and in case the null node
#     # is the root, reset the tree variable to a new TreeNode root.
#     if ancestor.is_root():
#         if dest.up == tree.treenode:
#             root = newnode
#         else:
#             root = [i for i in tree.treenode.children if i != ancestor][0]
#         tree.treenode.delete(True, True)
#     else:
#         root = tree.treenode
#     ancestor.delete(True, True)

#     # set src dist (optional allowing shrink to fit)
#     if src.height > dest.height + dist:
#         if not shrink:
#             raise ValueError(
#                 f"Clade {idx0} does not fit below clade {idx1}. "
#                 "Use shrink=True.\n"
#                 f"Details: node {idx0} height={src.height}; "
#                 f"node {idx1} height={dest.height + dist}; "
#                 f"dist={dist}; "
#                 f"Error b/c {src.height} > {dest.height} + {dist}")
#         orig = src.height + src.dist
#         scale = orig / (dest.height + dist)
#         for node in src.traverse():
#             node.dist /= scale
#     else:
#         src.dist = dest.height + dist - src.height

#     # set dest dist
#     dest.dist = dist

#     # connect newnode to its parent
#     dest_parent = dest.up
#     dest_parent.children.append(newnode)
#     dest_parent.children.remove(dest)

#     # connect newnode to its children
#     newnode.up = dest_parent
#     newnode.children = [src, dest]
#     src.up = newnode
#     dest.up = newnode

#     # re-toytree it from the root
#     tree = toytree.tree(root)

#     # TODO: add highlight boolean arg and color moved clade.
#     # tree.style.edge_colors = ['black'] * tree.nnodes
#     # for idx in tree.get_node_descendant_idxs(src)
#     return tree


def _resolve_nodes(
    node: Node,
    dist: float,
    support: float,
    rng: np.random.Generator,
    recursive: bool,
) -> Node:
    """Resolve multifurcating Nodes. Used in `resolve_polytomies`.

    Randomly choose left/right split among children.
    """
    if len(node.children) <= 2:
        return node.children

    # split the children into [left, (remaining)]
    child_idxs = range(len(node.children))
    idx = rng.choice(child_idxs)
    left = node.children[idx]
    remaining = [node.children[i] for i in child_idxs if i != idx]

    # create new right parent
    right = Node(dist=dist, support=support)

    # resolve this Node
    node._children = (left, right)
    left._up = node
    right._up = node

    # connect remaining as (bi-multi) of right
    for child in remaining:
        child._up = right
        right._add_child(child)

    # update children by recursively operating on right nodes
    if recursive:
        right._children = _resolve_nodes(
            right, dist, support, rng, recursive)
    return left, right


if __name__ == "__main__":

    import toytree
    toytree.set_log_level("DEBUG")
    t = toytree.rtree.unittree(16, treeheight=10)
    t.mod.prune(0, 1, 2, 3)
    t.draw()



#!/usr/bin/env python

"""Create Wright-Fisher plots.

This is used to demonstrate how genetic drift creates by allele
frequency change over generations, and generates genealogies.
"""

from typing import Optional, Dict
import toyplot
import numpy as np
from toytree import ToyTree, Node
from toytree.color import COLORS1, COLORS2, ToyColor, ColorType


class WrightFisherPlot:
    def __init__(self, time=20, popsize=20, seed: Optional[int] = None, **kwargs):
        self.rng: np.random.Generator = np.random.default_rng(seed)
        """: numpy random number generator."""
        self.grid: np.ndarray = None
        """: numpy int (time, 2N) array of node IDs as a grid."""
        self.edges: np.ndarray = None
        """: numpy int (2N * time, 2) array of edges by node IDs."""
        self.sampled_edges: np.ndarray = None
        """: numpy int (??, 2) array of edges for nsamples."""
        self.coords: np.ndarray = None
        """: numpy int (2N * time, 2) array of coordinates by ID."""
        self.canvas: toyplot.Canvas = None
        """: toyplot Canvas."""
        self.axes: toyplot.coordinates.Cartesian = None
        """: toyplot cartesian axes."""
        self.marks: Dict[str, 'toyplot.Mark'] = {}
        """: toyplot Marks."""
        self.time: int = time
        """: Number of discrete generations to simulate evolution."""
        self.popsize: int = popsize
        """: diploid population size (N)."""
        self.alleles: np.ndarray = np.zeros(popsize * 2 * time, dtype=int)
        """: numpy int array of allele states for gene copies by ID."""
        self._setup_grid_and_canvas(**kwargs)

    def _setup_grid_and_canvas(self, **kwargs) -> None:
        """Init values for grid, coords, canvas, and axes."""
        # get node coordinates
        self.grid = np.arange(self.popsize * 2 * self.time)
        self.grid = self.grid.reshape((self.time, self.popsize * 2))
        _ys, _xs = np.where(self.grid > -1)
        self.coords = np.column_stack([_xs, _ys[::-1]])

        # setup the canvas and axes
        self.canvas = toyplot.Canvas(
            width=kwargs.get("width", 500),
            height=kwargs.get("height", 500),
            style={"background-color": "white"},
        )
        self.axes = self.canvas.cartesian(
            ylabel="Time (generations)",
            xlabel="Gene copies",
            margin=70,
            padding=25,
        )
        self.axes.y.ticks.labels.style["font-size"] = 14
        self.axes.y.label.style["font-size"] = 16
        self.axes.y.label.offset = 30
        self.axes.y.ticks.locator = toyplot.locator.Extended()
        self.axes.y.ticks.show = True

        self.axes.x.ticks.show = False
        self.axes.x.ticks.labels.show = False
        self.axes.x.spine.show = False
        self.axes.x.label.style["font-size"] = 16
        self.axes.x.label.offset = 0

    def draw_diploids(self):
        """Adds a rectangle around pairs of gene copies to represent a diploid individual."""
        self.marks['diploids'] = self.axes.rectangle(
            self.coords[:, 0][::2] - 0.25,
            self.coords[:, 0][::2] + 1.25,
            self.coords[:, 1][::2] - 0.25,
            self.coords[:, 1][::2] + 0.25,
            style={
                "fill": "lightgrey", "stroke": "grey",
                "stroke-opacity": 0.75, "stroke-width": 1.5,
            },
        )

    def get_ancestry_edges(self, sort: bool=True) -> None:
        """Sample ancestor-descendant relationships and store to .edges.

        This defines the ancestry connecting all gene copies in each
        generation to an ancestor from the previous generation.
        """
        # iterate over each generation adding pairs of node indices
        for gen in range(self.grid.shape[0] - 1, 0, -1):

            # children idxs span from left to right
            lower_idxs = self.grid[gen]

            # randomly sample parent idxs (some have many children, some none)
            upper_idxs = self.rng.choice(
                self.grid[gen - 1], size=self.grid[gen - 1].size, replace=True)

            # get sorting index for the upper idxs
            if sort:
                order = np.argsort(upper_idxs)
            else:
                order = np.arange(upper_idxs.size)

            # update array of edges
            iedges = np.column_stack([lower_idxs, upper_idxs[order]])
            if self.edges is not None:
                self.edges = np.concatenate([self.edges, iedges])
            else:
                self.edges = iedges

    def draw_edges(self, **kwargs):
        """Draw ancestry edges with information in .edges."""
        # style the graph
        style = {
            "vlshow": False, "ecolor": "black", "ewidth": 1.25,
            'vsize': 0, 'estyle': {}, 'eopacity': 0.6}
        style.update(kwargs)

        # add graph lines from lower_idxs to upper_idxs, using coordinates for all
        self.marks['genealogy'] = self.axes.graph(
            self.edges,
            vcoordinates=self.coords[sorted(np.unique(self.edges))],
            **style,
        )

    def get_sampled_edges(self, nsamples: int):
        """Add sampled genealogy lines from N random samples, or a list of sampled indices."""
        if not nsamples:
            return
        samples = list(self.rng.choice(self.grid[-1], size=nsamples, replace=False))
        tracked = samples.copy()
        for edg in self.edges:
            if edg[0] in tracked:
                tracked.append(edg[1])

        # get the node pairs that make up each of these edges
        mask = np.isin(self.edges[:, 0], tracked)
        self.sampled_edges = self.edges[mask]

    def draw_sampled_edges(self, **kwargs):
        """Draw genealogy connecting sampled gene copies."""
        style = {
            "vlshow": False,
            "vsize": 8,
            "vmarker": 'o',
            "vcolor": "black",
            "vstyle": {'stroke': 'black', 'stroke-width': 2},
            "ecolor": 'black',
            "ewidth": 3,
            "estyle": {},
        }
        style.update(kwargs)
        mark = self.axes.graph(
            self.sampled_edges,
            vcoordinates=self.coords[sorted(np.unique(self.sampled_edges))],
            **style,
        )
        self.marks['genealogy'] = mark

    def get_allele_states(self, allele_frequency=0.5):
        """Fills allele states based on ancestry edges."""
        if allele_frequency in [0, 1]:
            self.alleles[:] = 0
        else:
            self.alleles[:2 * self.popsize] = self.rng.binomial(
                n=1, p=allele_frequency, size=2 * self.popsize)
            for child in range(self.popsize * 2, self.popsize * 2 * self.time):
                parent = self.edges[self.edges[:, 0] == child, 1][0]
                self.alleles[child] = self.alleles[parent]

    def draw_nodes(self, color: Optional[ColorType], **kwargs):
        """Add circle marks for gene copies showing allele frequencies.

        Allele frequencies are mapped on generation 1 randomly and then
        inherited in all following generations based on genealogy edges.

        Parameters
        ----------
        p: float
            Frequency of allele 1 at a mono- or bi-allelic locus.
        """
        style = {
            "marker": "o",
            "size": 6,  #self.canvas.width / self.grid.shape[1] / 7.5,
            "mstyle": {
                "stroke": "black",
                "stroke-opacity": 0.75,
                "stroke-width": 1.5,
            },
        }
        style.update(kwargs)

        if color is None:
            color1 = ToyColor(COLORS1[0]).css
            color2 = ToyColor(COLORS1[1]).css
        else:
            color = ToyColor(color)
            rgb = [1 - i for i in color.rgba[:3]]
            color1 = color.css
            color2 = ToyColor(tuple(rgb + [1.0])).css
        style['color'] = [color1 if i == 0 else color2 for i in self.alleles]

        # plot marks
        mark = self.axes.scatterplot(
            self.coords[:, 0], self.coords[:, 1],
            **style,
        )
        self.marks['haploids'] = mark

    def get_genealogy(self, nsamples: int) -> ToyTree:
        """Returns a toytree matching to the evolved genealogy."""

        raise NotImplementedError("Coming soon.")
        # build dict of samples {0: Node(), 1: Node(), 2: Node(), ...}
        nodes = {i: Node(name=i) for i in range(nsamples)}

        # build coal time dict: {time0: (0, 1), time1: (2, 3), ...}
        # ...

        # build tree connecting nodes at coal times.
        tree = toytree.ToyTree()


def wright_fisher_simulation(
    time: int,
    popsize: int,
    seed: Optional[int] = None,
    nsamples: Optional[int] = None,
    sort_edges: bool = True,
    show_diploids: bool = False,
    allele_frequency: float = 1.0,
    node_size: int = 6,
    node_color: Optional[ColorType] = None,
    **kwargs,
) -> toyplot.Canvas:
    """Simulate a population evolving as a Wright-Fisher process.

    A WF process model generates 2N haploid gene copies in discrete
    non-overlapping generations, where each generation gene copies
    are randomly sampled with replacement from the previous generation.
    This process creates a genealogy of edges connecting all gene
    copies. Some gene copies leave no descendants, while others can
    leave multiple. Genetic variation, represented by different
    alleles (e.g., colors) among gene copies will be lost over time
    by this random sampling procedure, representing a manifestation
    of genetic drift.

    Parameters
    ----------
    time: int
        Number of generations to simulate evolution.
    popsize: int
        The number of diploid individuals in the population (N).
    seed: int
        Optional random number seed to make simulation repeatable.
    nsamples: int
        Optional, number of random gene copies to sample at the
        present. A genealogy will be drawn to connect the samples.
    sort_edges: bool
        If True the gene copies will be sorted so that edges are
        easier to interpret.
    show_diploids: bool
        If True a rectangle is drawn around pairs of gene copies to
        show how they can be randomly joined into diploid genomes.
    node_size: int
        The size of node markers for gene copies.
    node_color: Color
        Optional, a color of node markers for gene copies. If None
        then
    allele_frequency: float
        Starting allele frequency for bi-allelic variation.
    """
    model = WrightFisherPlot(seed=seed, time=time, popsize=popsize, **kwargs)
    model.get_ancestry_edges(sort=sort_edges)
    model.get_allele_states(allele_frequency=allele_frequency)
    model.get_sampled_edges(nsamples=nsamples)

    if show_diploids:
        model.draw_diploids()
    model.draw_edges()
    if nsamples is not None:
        model.draw_sampled_edges(vsize=node_size + 2)
    model.draw_nodes(size=node_size, color=node_color)
    return model


if __name__ == "__main__":

    import toyplot.browser

    # wf = WrightFisherPlot(seed=None, time=40, popsize=16, width=500, height=800)
    # # wf.add_diploids()
    # wf.add_lines(sort=True)
    # wf.add_sampled_lines(6)
    # wf.add_haploids()            # this should go last

    wf = wright_fisher_simulation(
        time=30, popsize=16, seed=123, nsamples=5,
        width=500, height=500,
        show_diploids=True,
        node_size=6,
        allele_frequency=0.5,
        # node_color='red'
    )
    toyplot.browser.show(wf.canvas)



#!/usr/bin/env python

"""Implementatin of supertree inference methods.



References
----------
https://pubmed.ncbi.nlm.nih.gov/39119174/
"""

import numpy as np
from pandas import DataFrame
from toytree.core import ToyTree
from toytree.core.multitree import MultiTree


def get_matrix_representation(tree: ToyTree, tip_labels: list[str], df: bool = False) -> np.ndarray:
    """Return a matrix representation of a set of rooted trees.

    This matrix can be analyzed using a phylogenetic inference tool
    to infer a tree. A common approach is to infer a parsimony tree in
    which case the supertree is termed a matrix representation with
    parsimony (MRP) tree.

    Parameters
    ----------
    trees: ToyTree
        ...
    tip_labels: list[str]
        ...
    df: bool
        Return table as a pandas DataFrame with tip_labels as the index
        and columns corresponding to tree
    """
    # get array filled w/ 1 for tips in each clade
    data = []
    for bipart in tree.enum.iter_bipartitions():
        dat = np.zeros(len(tip_labels), dtype=int)
        for i in bipart[0]:
            dat[tip_labels.index(i)] = 1
        data.append(dat)

    # get mask of missing taxa in this tree
    mask = np.zeros(len(tip_labels), dtype=np.bool_)
    for i in set(tip_labels) - set(tree.get_tip_labels()):
        mask[tip_labels.index(i)] = True

    # create a masked array to
    marr = np.ma.array(
        data=data,
        mask=[mask for i in data],
        dtype=int,
    )

    # convert to dataframe
    if df:
        return DataFrame(marr.T, index=tip_labels)
    print(marr.T)


if __name__ == "__main__":

    import toytree
    t0 = "((a,b),((c,d),f));"
    t1 = "((a,b),((c,d),e));"
    t2 = "((a,c),((b,d),(e,f)));"
    t3 = "((a,c),(b,e));"

    names = list("abcdef")
    trees = toytree.mtree([t0, t1, t2, t3])
    for tree in trees:
        marr = get_matrix_representation(tree, names, df=True)
        print(marr)



#!/usr/bin/env python

"""Functions for `distance` based calculations in toytree. 

Distances can be measured between Nodes of the same ToyTree, or among
different ToyTrees. The first represents either the number of Nodes 
separating two or more Nodes, or the sum of edge lengths separating 
them. Tree distance measurements can use topologies and/or edge 
lengths. Each can be returned for individual pairs, or as a table
with pairwise distances among many objects.

These functions are accessible either at (:mod:`toytree.distance`) 
subpackage, or in the tree-level API at (:mod:`ToyTree.distance`).

Examples
--------
>>> tree1 = toytree.rtree.unittree(ntips=10, seed=123)
>>> tree2 = toytree.rtree.unittree(ntips=10, seed=321)

>>> # get distances between Nodes of the same tree
>>> dist_01 = toytree.distance.get_node_distance(tree1, 0, 1)
>>> dists = toytree.distance.get_node_distance_matrix(tree1)
>>> dists_1234 = toytree.distance.get_node_distance_matrix(tree1, 1, 2, 3, 4)

>>> # get distance between ToyTrees
>>> dist_t01 = toytree.distance.get_tree_distance_qrt(tree1, tree2)
>>> dist_t01 = toytree.distance.get_tree_distance_rf(tree1, tree2)

>>> # tree API usage (funcs accessible from trees for convenience)
>>> tree.distance.get_node_distance
>>> tree.distance.get_node_distance_matrix
"""

# Note
# -----

from toytree.distance._src.nodedist import *
from toytree.distance._src.treedist import *
from toytree.distance._src.quartet_dist import *



#!/usr/bin/env python

"""toytree inference subpackage.


"""

from toytree.infer.src.upgma import infer_upgma_tree
from toytree.infer.src.neighbor_joining import infer_neighbor_joining_tree
from toytree.infer.src.consensus import get_consensus_tree, get_consensus_features

# requires sympy which is not yet in conda recipe, so for now
# you need to call the following to access the likelihood code:
# >>> from toytree.infer.src import maximum_likelihood
# >>> from toytree.infer.src.maximum_likelihood import JC69, K80, TN93, get_tree_likelihood, get_tree_likelihood_plot_gen




#!/usr/bin/env python

"""Validate style arguments to ToyTree.draw.

Note: validate_x() functions are written specifically to work in two
fairly different scenarios: `draw_tree` and `annotation`. 

In `draw_tree` they take an attribute of a TreeStyle as input, expand
it to an array, and return it to set it back on the TreeStyle object.
It could be simpler to just set values on the TreeStyle, but for
compatibility with the second method we don't do that.

on the dict info
of TreeStyle objects because this allows for re-using these more
easily in annotation functions.
"""

from typing import Mapping, Any, TypeVar
from loguru import logger

from toytree.style import TreeStyle

from toytree.style.src.validate_data import (
    validate_numeric,
    validate_mask,
    validate_markers,
    validate_colors,
    validate_hover,
    validate_labels,
    validate_admixture_edges,
)
from toytree.style.src.validate_nodes import validate_node_style
from toytree.style.src.validate_node_labels import validate_node_labels_style
from toytree.style.src.validate_tips import validate_tip_labels_style
from toytree.style.src.validate_edges import (
    validate_edge_style,
    validate_edge_align_style,
)

# from toytree.style.src.validate_edges import validate_edge_widths
logger = logger.bind(name="toytree")
ToyTree = TypeVar("ToyTree")
NON_VALIDATED = [
    "height",
    "width",
    "layout",
    "edge_type",
    "tip_labels_align",
    "use_edge_lengths",
    "scale_bar",
    "padding",
    "xbaseline",
    "ybaseline",
    "shrink",
    "node_as_edge_data",
    # "admixture_edges",
    # "fixed_order",
    # "fixed_position",
]


def get_dict(kwargs: Mapping[str, Any], name: str) -> Any:
    """Return user value if provided (not None) else BaseStyle value.

    This is necessary because the key is usually present and mapped to
    a value of None, but we want {} returned if the value is None.
    """
    return kwargs[name] if kwargs.get(name) is not None else {}


def validate_style(tree: ToyTree, style: TreeStyle, **kwargs) -> TreeStyle:
    """Validate style arguments to ToyTree.draw() before creating Mark.

    The TreeStyle is a copy from the ToyTree and kwargs arguments are
    user args to the .draw() function which override the TreeStyle
    defaults. This expands arguments into values and checks the size
    and type. Many of the functions within are also used in annotations.

    Note
    ----
    This function is only used in `ToyTree.draw`. By contrast,
    `toytree.annotation` only calls the validation funcs within this.

    Parameters
    ----------
    tree: ToyTree
        ...
    style: TreeStyle
        A base or modified TreeStyle dict.
    **kwargs: Dict[str, Any]
        A dict with any user-supplied style arguments to draw_toytree.
    """
    # validate node settings
    # style.node_mask = validate_node_mask(tree, style, **kwargs)
    # style.node_sizes = validate_node_sizes(tree, style, **kwargs)
    # style.node_markers = validate_node_markers(tree, style, **kwargs)
    # style.node_hover = validate_node_hover(tree, style, **kwargs)
    # style.node_labels = validate_node_labels(tree, style, **kwargs)
    # style.node_colors, node_fill_color = validate_node_colors(tree, style, **kwargs)
    # style.edge_widths = validate_edge_widths(tree, style, **kwargs)
    # style.edge_colors, edge_stroke = validate_edge_colors(tree, style, **kwargs)
    # style.tip_labels = validate_tip_labels(tree, style, **kwargs)
    # style.tip_labels_angles = validate_tip_labels_angles(tree, style, **kwargs)
    # style.tip_labels_colors, tip_fill_color = validate_tip_labels_colors(tree, style, **kwargs)
    style.node_mask = validate_mask(
        tree, tree_style=style, style=kwargs)
    style.node_sizes = validate_numeric(
        tree, key="node_sizes", size=tree.nnodes, tree_style=style, style=kwargs)
    style.node_markers = validate_markers(
        tree, key="node_markers", size=tree.nnodes, tree_style=style, style=kwargs)
    style.node_hover = validate_hover(
        tree, key="node_hover", size=tree.nnodes, tree_style=style, style=kwargs)
    style.node_labels = validate_labels(
        tree, key="node_labels", size=tree.nnodes, tree_style=style, style=kwargs)
    style.node_colors, node_fill_color = validate_colors(
        tree, key="node_colors", size=tree.nnodes, tree_style=style, style=kwargs)
    if node_fill_color is not None:
        style.node_style.fill = node_fill_color

    # style.edge_markers = validate_markers(
    #     tree, key="edge_markers", size=tree.nnodes - 2, tree_style=style, style=kwargs)
    # style.edge_labels = validate_labels(
    #     tree, key="edge_labels", size=tree.nnodes - 2, tree_style=style, style=kwargs)
    # edge_hover: TODO

    # validate edge settings
    style.edge_widths = validate_numeric(
        tree, key="edge_widths", size=tree.nnodes, tree_style=style, style=kwargs)
    style.edge_colors, edge_stroke = validate_colors(
        tree, key="edge_colors", size=tree.nnodes, tree_style=style, style=kwargs)
    if edge_stroke is not None:
        style.edge_style.stroke = edge_stroke

    # validate tip settings. Override tip_labels=None with tip_labels="name"
    style.tip_labels = validate_labels(
        tree, key="tip_labels", size=tree.ntips, tree_style=style, style=kwargs)
    style.tip_labels_angles = validate_numeric(
        tree, key="tip_labels_angles", size=tree.ntips, tree_style=style, style=kwargs)
    style.tip_labels_colors, tip_fill_color = validate_colors(
        tree, key="tip_labels_colors", size=tree.ntips, tree_style=style, style=kwargs)
    if tip_fill_color is not None:
        style.tip_labels_style.fill = tip_fill_color

    # validate style dictionaries
    style.node_style = validate_node_style(tree, style.node_style, **get_dict(kwargs, "node_style"))
    style.edge_style = validate_edge_style(tree, style.edge_style, **get_dict(kwargs, "edge_style"))
    style.node_labels_style = validate_node_labels_style(tree, style.node_labels_style, **get_dict(kwargs, "node_labels_style"))
    style.edge_align_style = validate_edge_align_style(tree, style.edge_align_style, **get_dict(kwargs, "edge_align_style"))
    style.tip_labels_style = validate_tip_labels_style(tree, style.tip_labels_style, **get_dict(kwargs, "tip_labels_style"))

    style.admixture_edges = validate_admixture_edges(tree, style=kwargs)

    # set remaining key:val that do not have validators
    for key in NON_VALIDATED:
        val = kwargs.get(key, None)
        if val is not None:
            setattr(style, key, val)
    # for key, val in kwargs.items():
    #     if key in NON_VALIDATED:
    #         if val is not None:
    #             setattr(style, key, val)
    return style


if __name__ == "__main__":

    import toytree
    # from toytree.style import get_base_tree_style_by_name
    from toytree.drawing.src.draw_toytree import get_tree_style_base

    tree = toytree.rtree.unittree(8)

    style = get_tree_style_base(tree, ts='c', node_colors="red")
    style = validate_style(tree, style)

    print(style.node_colors)
    print(style.node_style)

    tree.draw(node_mask=False)
    tree._draw_browser(tmpdir="~")



#!/usr/bin/env python

"""Function to split or combine color arguments for broader support.

Many tools do not support color as (r,g,b,a) but instead take (r,g,b)
and opacity as a separate style arg. This function is used to split
ToyColor objects to separate the opacity.
"""

from typing import Dict
# from loguru import logger
from toytree.color.src.toycolor import ToyColor

# logger = logger.bind(name="toytree")


# def concat_style_to_str(style: Dict[str, str]) -> str:
#     """Return a style dict concatenated into a style string.

#     Example
#     -------
#     >>> x = {'fill': 'rgb(100%,0%,0%)', 'fill-opacity': 1.0}
#     >>> print(style_to_string(x))
#     >>> # 'fill:rgb(100%,0%,0%);fill-opacity:1.0'
#     """
#     strs = [f"{key}:{value}" for key, value in sorted(style.items())]
#     return ";".join(strs)


# def color_style_split(style: Dict[str, Any]) -> Dict[str, Any]:
#     """

#     This is used in annotation functions to fix up colors in style
#     dicts before passing to toyplot drawing functions.

#     Examples
#     --------
#     >>> d = {'fill': ToyColor("red"), fill-opacity: 0.5}
#     >>> color_style_split(d)
#     >>> # {'fill': (1, 0, 0), 'fill-opacity': 0.5}

#     >>> d = {'fill': ToyColor((1, 0, 0, 0.5))}
#     >>> color_style_split(d)
#     >>> # {'fill': (1, 0, 0), 'fill-opacity': 0.5}
#     """
#     if not style:
#         return {}

#     fill = style.pop("fill", None)
#     fillo = style.pop("fill-opacity", None)
#     stroke = style.pop("stroke", None)
#     strokeo = style.pop("stroke-opacity", None)

#     # get fill or stroke style string with opacity separated and optinally
#     # overwritten if provided as an argument. The 'if not X' arg here
#     # will return False for transparent color (0,0,0,0) which we use up
#     # until this point for missing data (np.nan) colors.
#     styles = []

#     # if None then use parent value
#     if fill is not None:
#         # if (0,0,0,0) then suppress style
#         if not fill:
#             styles.append("fill:none")
#         else:
#             styles.append(ToyColor(fill).get_style_str(opacity=fillo))

#     # if None then use parent value
#     if stroke is not None:
#         # if (0,0,0,0) then suppress style
#         if not stroke:
#             styles.append("stroke:none")
#         else:
#             styles.append(ToyColor(stroke).get_style_str(stroke=True, opacity=strokeo))

#     # keep any other styles in style dict in the style string
#     for key, value in sorted(style.items()):
#         if value is not None:
#             styles.append(f"{key}:{value}")


def concat_style_fix_color(style: Dict[str, str], extra: str = None) -> str:
    """Return a style dict concatenated to string w/ color fixes.

    The style dict here will have fill/stroke color and opacity items
    that need to be evaluated. The colors will always be np.void
    type, representing a single color from an array of colors.

    Note
    ----
    This is destructive to the style dict, popping items from it.
    """
    if not style:
        return ""
    style = style.copy()  # for now, needed for node.style.fill-opacity
    fill = style.pop("fill", None)
    fillo = style.pop("fill-opacity", None)
    stroke = style.pop("stroke", None)
    strokeo = style.pop("stroke-opacity", None)

    # get fill or stroke style string with opacity separated and optinally
    # overwritten if provided as an argument. The 'if not X' arg here
    # will return False for transparent color (0,0,0,0) which we use up
    # until this point for missing data (np.nan) colors.
    styles = []

    # if None then use parent value
    if fill is not None:
        # if (0,0,0,0) then suppress style
        if not fill:
            styles.append("fill:none")
        else:
            styles.append(ToyColor(fill).get_style_str(opacity=fillo))

    # if None then use parent value
    if stroke is not None:
        # if (0,0,0,0) then suppress style
        if not stroke:
            styles.append("stroke:none")
        else:
            styles.append(ToyColor(stroke).get_style_str(stroke=True, opacity=strokeo))

    # keep any other styles in style dict in the style string
    for key, value in sorted(style.items()):
        if value is not None:
            styles.append(f"{key}:{value}")

    # add additional string styles
    if extra:
        styles.append(extra)
    return ";".join(styles)


if __name__ == "__main__":

    # print("result", concat_style_fix_color({"fill": (0, 0, 0, 0)}))
    # print("result", concat_style_fix_color({"fill": (0., 0., 0., 0.)}))
    sty1 = {
        "fill": ToyColor((0., 0., 0., 0.5)),
        "fill-opacity": False,
    }
    print(f"result: '{concat_style_fix_color(sty1)}'")

    sty2 = {
        "fill": ToyColor((0., 0., 0., 0.5)),
        "fill-opacity": None,
    }
    print(f"result: '{concat_style_fix_color(sty2)}'")

    sty3 = {
        "fill": ToyColor((0., 0., 0., 0.5)),
        "fill-opacity": 1.0,
    }
    print(f"result: '{concat_style_fix_color(sty3)}'")

    sty4 = {
        "stroke": ToyColor((0., 0., 0., 0.5)),
        "stroke-opacity": 0.1,
    }
    print(f"result: '{concat_style_fix_color(sty4)}'")

    sty5 = {
        "stroke": ToyColor((0., 0., 0., 0.5)),
        "stroke-opacity": False,
    }
    print(f"result: '{concat_style_fix_color(sty5)}'")

    sty6 = {
        "stroke": ToyColor((0., 0., 0., 0.5)),
        "stroke-opacity": None,
    }
    print(f"result: '{concat_style_fix_color(sty6)}'")

    sty7 = {
        "stroke": ToyColor((0., 0., 0., 0.5)),
        "stroke-opacity": 1.0,
    }
    print(f"result: '{concat_style_fix_color(sty7)}'")

    sty8 = {
        "fill": ToyColor((0., 0., 0., 0.5)),
        "fill-opacity": 1.0,
        "stroke": ToyColor((0., 0., 0., 0.5)),
        "stroke-opacity": 1.0,
    }
    print(f"result: '{concat_style_fix_color(sty8)}'")

    sty9 = {
        "fill": ToyColor((0., 0., 0., 0.5)),
        "fill-opacity": 1.0,
        "stroke": ToyColor((0., 0., 0., 0.5)),
        "stroke-opacity": 1.0,
        "stroke-width": 3.0,
    }
    print(f"result: '{concat_style_fix_color(sty9)}'")



#!/usr/bin/env python

"""Node distance functions.

These functions take a tree as input and use Node int idx labels
to select Nodes.
"""

from typing import TypeVar, Tuple, Union, Dict, Iterator
import itertools
import numpy as np
import pandas as pd
from loguru import logger
from toytree import Node, ToyTree
from toytree.core.apis import TreeDistanceAPI, add_subpackage_method
# from toytree.utils import ToytreeError

# type aliases
Query = TypeVar("Query", str, int, Node, None)
logger = logger.bind(name="toytree")

# put functions here to have then exposed to 'distance' subpackage API
__all__ = [
    "get_node_path",
    "iter_node_path",
    "get_node_distance",
    "get_node_distance_matrix",
    "get_internal_node_distance_matrix",
    "get_tip_distance_matrix",
    "get_descendant_dists",
    "iter_descendant_dists",
    "get_farthest_node",
    "get_farthest_node_distance",
]


@add_subpackage_method(TreeDistanceAPI)
def get_node_path(tree: ToyTree, node0: Query, node1: Query) -> Tuple[Node]:
    """Return a list of Nodes connecting two Nodes (including at ends).

    Parameters
    ----------
    tree: ToyTree
        A tree containing the queried Nodes.
    node0: int, str, or Node
        A Node in the tree at the start of the path. Nodes can be
        queried by int idx label, str name, or as Node objects.
    node1: int, str, or Node
        A Node in the tree at the end of the path.

    Example
    -------
    >>> tree = toytree.rtree.imbtree(10)
    >>> tree.distance.get_node_path(0, 2)
    >>> # (<Node(idx=0)>, <Node(idx=10)>, <Node(idx=11)>, <Node(idx=2)>)
    """
    return tuple(iter_node_path(tree, node0, node1))


@add_subpackage_method(TreeDistanceAPI)
def iter_node_path(tree: ToyTree, node0: Query, node1: Query) -> Iterator[Node]:
    """Generator of the path between two Nodes."""
    n0 = tree.get_nodes(node0)[0]
    n1 = tree.get_nodes(node1)[0]
    mrca = tree.get_mrca_node(n0, n1)

    # yield start of path
    if n0 != mrca:
        yield n0

    # yield from start -> mrca
    for node in n0.iter_ancestors(root=mrca):
        yield node

    # yield mrca
    yield mrca

    # yield path from mrca -> end
    for node in list(n1.iter_ancestors(root=mrca))[::-1]:
        yield node

    # yield end
    if n1 != mrca:
        yield n1


# >3X faster than older TreeNode.get_distance(), and scales better.
# If the tree is ultrametric we could calculate 2X faster by just
# doubling the distance to mrca...
@add_subpackage_method(TreeDistanceAPI)
def get_node_distance(
    tree: ToyTree,
    node0: Query,
    node1: Query,
    topology_only: bool = False,
) -> float:
    """Return patristic distance between two Nodes on a ToyTree.

    Parameters
    ----------
    tree: toytree.ToyTree
        A ToyTree instance.
    node0: int, str, or Node
        A Node in the tree.
    node1: int, str, or Node
        Another Node in the tree.
    topology_only: bool
        If True then distance is measured as number of edges between
        the two Nodes.

    Returns
    -------
    distance: float
        The patristic distance between two nodes (the minimum
        distance along edges connecting them that passes through
        their common ancestor).

    Example
    -------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> toytree.distance.get_node_distance(tree, 0, 1)
    """
    # get query as Nodes (order not maintained, but not needed)
    nodes = tree.get_nodes(node0, node1)
    if len(nodes) == 1:
        return 0
    if len(nodes) > 2:
        q0 = tree.get_nodes(node0)
        q1 = tree.get_nodes(node1)
        msg = f"Bad Node queries: node0 matched {q0}; node1 matched {q1}"
        logger.error(msg)
        raise ValueError(msg)
    node0, node1 = nodes

    # get mrca of the query
    mrca = tree.get_mrca_node(node0, node1)

    # store total distance
    dist = 0

    # count from each node up to mrca
    for node in (node0, node1):
        # skip if node == mrca
        if node != mrca:
            # get 1 or dist for every node up to the mrca
            if topology_only:
                dist += 1 + sum(1 for i in node.iter_ancestors(mrca))
            else:
                dist += node._dist + sum(i._dist for i in node.iter_ancestors(mrca))
    return dist


@add_subpackage_method(TreeDistanceAPI)
def get_node_distance_matrix(
    tree: ToyTree,
    topology_only: bool = False,
    df: bool = False
) -> Union[np.array, pd.DataFrame]:
    """Return pairwise distances between all Nodes in a ToyTree.

    Parameters
    ----------
    tree: toytree.ToyTree
        A ToyTree.
    topology_only: bool
        If True distances represent the number of edges between Nodes.
    df: bool
        If True a pandas.DataFrame is returned instead of np.ndarray.

    Returns
    -------
    node_distance_matrix: np.ndarray or pd.DataFrame
        A matrix is returned as a np.ndarray with rows and columns
        ordered by Node int idx labels, or as a pd.DataFrame with
        row and column names as str Node names for leaf Nodes and idx
        labels for internal Nodes.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> toytree.distance.get_node_distance_matrix(tree)
    """
    dtype = int if topology_only else float

    # postorder ordered
    arr = np.zeros((tree.nnodes, tree.nnodes), dtype=dtype)

    # store list of descendants indices for each node
    clade_map = {}

    # reorder
    reorder = []

    # traverse tree in postorder, but first visit all tips (idxorder).
    for idx, node in enumerate(tree.traverse("postorder")):

        # init clade list to store descendants of this node
        clade_map[node] = [idx]

        # store idxorder index for re-sorting later.
        reorder.append(node.idx)

        # build desc list from children desc lists
        for child in node._children:

            # store child's clade list to parent's clade list
            clade = clade_map[child]
            clade_map[node].extend(clade)

            # set dist from child's clade members to parent
            arr[clade, idx] = arr[clade, max(clade)] + (1 if topology_only else child._dist)

        # set children clade's dists to each other
        for ch0, ch1 in itertools.combinations(node._children, 2):
            ch0 = clade_map[ch0]
            ch1 = clade_map[ch1]
            for c in ch0:
                arr[c, ch1] = arr[c, idx] + arr[ch1, idx]

    # mirror fill
    arr[np.tril_indices_from(arr)] = arr.T[np.tril_indices_from(arr)]

    # reorder into idx order
    idxorder = [reorder.index(i) for i in range(tree.nnodes)]

    # optionally format as dataframe
    if not df:
        return arr[idxorder][:, idxorder]
    index = tree.get_tip_labels() + [str(i.idx) for i in tree[tree.ntips:]]
    return pd.DataFrame(
        arr[idxorder][:, idxorder], columns=index, index=index)


@add_subpackage_method(TreeDistanceAPI)
def get_internal_node_distance_matrix(
    tree: ToyTree,
    topology_only: bool = False,
    df: bool = False,
) -> Union[np.array, pd.DataFrame]:
    """Return pairwise distances between non-leaf Nodes in a ToyTree.

    Parameters
    ----------
    tree: toytree.ToyTree
        A ToyTree.
    topology_only: bool
        If True distances represent the number of edges between Nodes.
    df: bool
        If True a pandas.DataFrame is returned instead of np.ndarray.

    Returns
    -------
    node_distance_matrix: np.ndarray or pd.DataFrame
        A matrix is returned as a np.ndarray with rows and columns
        ordered by Node int idx labels, or as a pd.DataFrame.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> toytree.distance.get_internal_node_distance_matrix(tree)
    """
    arr = get_node_distance_matrix(tree, topology_only=topology_only, df=df)
    if not df:
        return arr[tree.ntips:, tree.ntips:]
    return arr.iloc[tree.ntips:, tree.ntips:]


@add_subpackage_method(TreeDistanceAPI)
def get_tip_distance_matrix(
    tree: ToyTree,
    topology_only: bool = False,
    df: bool = False
) -> Union[np.array, pd.DataFrame]:
    """Return pairwise distances between tip Nodes in a ToyTree.

    Parameters
    ----------
    tree: toytree.ToyTree
        The input ToyTree instance.
    topology_only: bool
        If True then all edges lengths are set to 1.
    df: bool
        If True a pandas.DataFrame is returned instead of np.array
        with str Node names as index and column names.

    Returns
    -------
    node_distance_matrix: np.ndarray or pd.DataFrame
        A matrix is returned as a np.ndarray with rows and columns
        ordered by Node int idx labels, or as a pd.DataFrame with
        row and column names as str Node names.

    Examples
    --------
    >>> tree = toytree.rtree.unittree(10, seed=123)
    >>> toytree.distance.get_tip_distance_matrix(tree)
    """
    arr = get_node_distance_matrix(tree, topology_only=topology_only, df=df)
    if not df:
        return arr[:tree.ntips, :tree.ntips]
    return arr.iloc[:tree.ntips, :tree.ntips]


@add_subpackage_method(TreeDistanceAPI)
def get_descendant_dists(
    tree: ToyTree,
    node: Query = None,
    topology_only: bool = False,
) -> Dict[Node, float]:
    """Return a dict {Node: distance} for Nodes descended from a
    selected Node (default=root).

    Distance is measured by the sum of edge lengths separating them,
    unless `topology_only` is True, in which case it is the number of
    edges separating them.

    This is function is a bit faster than calculating the full distance
    matrix when you only need distances to descendants of a Node. Values
    are generated in "preorder" traversal order (left then right).

    Example
    -------
    >>> tree = toytree.rtree.unittree(5, seed=123)
    >>> ndict = _get_dist_to_descendants_dict(tree)
    >>> # {Node(idx=0): 1.0, Node(idx=1): 2.0, ...}
    """
    # get distances among descendant nodes by traversal
    return dict(iter_descendant_dists(tree, node, topology_only))


@add_subpackage_method(TreeDistanceAPI)
def iter_descendant_dists(
    tree: ToyTree,
    node: Query = None,
    topology_only: bool = False,
) -> Iterator[Tuple[Node, float]]:
    """Generator of (Node, dist) tuples of descendant Nodes from a
    selected Node (default=root).

    Distance is measured as the sum of edge lengths between Nodes,
    unless `topology_only` is True, in which case it is the number of
    edges separating them.

    This is function is a bit faster than calculating the full distance
    matrix when you only need distances to descendants of a Node. Values
    are generated in "preorder" traversal order (left then right).

    Example
    -------
    >>> tree = toytree.rtree.unittree(5, seed=123)
    >>> list(tree.distance.iter_descendants_dists())
    >>> # [(Node(idx=0), 1.0), (Node(idx=1, 2.0), ...]
    """
    # get distances among descendant nodes by traversal
    node = tree.treenode if node is None else tree.get_nodes(node)[0]
    ndists = {i._idx: 0 for i in tree}
    for desc in node.traverse("preorder"):
        if desc != node:
            ndists[desc._idx] += 1 if topology_only else desc.dist
            if desc._up._idx in ndists:
                ndists[desc._idx] += ndists[desc._up._idx]
        yield desc, ndists[desc._idx]


@add_subpackage_method(TreeDistanceAPI)
def get_farthest_node(
    tree: ToyTree,
    node: Query,
    topology_only: bool = False,
    descendants_only: bool = False,
) -> Node:
    """Return the farthest Node from a selected Node.

    Note
    ----
    If >1 Nodes are equally distant the one w/ lowest idx is returned.
    See `get_node_distance_matrix` or `iter_descendant_dists` for other
    options to visit multiple equally distant Nodes.

    Parameters
    ----------
    tree: ToyTree
        The ToyTree on which to measure Node distances.
    node: int, str, or Node
        Node in the tree selected by int idx label, str name, or Node.
    topology_only: bool
        If True distances are measured as number of edges between
        Nodes rather than by the sum of edge distances.
    descendants_only: bool
        If True then the farthest Node is only searched among the
        descendants of the 'node' query, or the root is no Node was
        selected. If False then the farthest Node is searched across
        the entire tree using `get_node_distance_matrix()`.
    """
    # get distances to all, or only descendants
    node = tree.treenode if node is None else tree.get_nodes(node)[0]
    if descendants_only:
        ndists = get_descendant_dists(tree, node, topology_only)
        return max(ndists, key=lambda x: ndists[x])
    else:
        ndists = get_node_distance_matrix(tree, topology_only)
        nidx = ndists[node.idx].argmax()
        return tree[nidx]


@add_subpackage_method(TreeDistanceAPI)
def get_farthest_node_distance(
    tree: ToyTree,
    node: Query = None,
    topology_only: bool = False,
    descendants_only: bool = False,
) -> float:
    """Return distance to the farthest Node from a selected Node.

    Parameters
    ----------
    tree: ToyTree
        The ToyTree on which to measure Node distances.
    idx: int
        The int idx label of a Node from which to measure distances
        from. Default is None, which uses the root Node.
    topology_only: bool
        If True distances are measured as number of edges between
        two Nodes, rather than the sum of edge distances.
    descendants_only: bool
        If True then the farthest descendant Node is returned, rather
        than the farthest Node spanning any path on the tree.
    """
    node = tree.treenode if node is None else tree.get_nodes(node)[0]
    if descendants_only:
        return max(i[1] for i in iter_descendant_dists(tree, node, topology_only))
    return max(get_node_distance_matrix(tree, topology_only)[node._idx])


if __name__ == "__main__":

    import toytree
    # TREE = toytree.rtree.unittree(10, seed=123)
    # print(TREE.draw())
    # print(get_node_distance_matrix(TREE, True))

    TREE = toytree.rtree.unittree(5, seed=123)
    print(TREE.distance.get_farthest_node(5))
    print(get_node_path(TREE, 0, 5))
    TREE.treenode.draw_ascii()

    print(get_node_distance_matrix(TREE, df=True))
    print(get_tip_distance_matrix(TREE, df=True))
